# Copyright (C) 2009 by Eric Talevich (eric.talevich@gmail.com)
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Classes corresponding to phyloXML elements.

See Also
--------
Official specification:
   http://phyloxml.org/
Journal article:
    Han and Zmasek (2009), https://doi.org/10.1186/1471-2105-10-356

"""

import re
import warnings

from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq
from Bio.SeqFeature import SeqFeature, FeatureLocation
from Bio.SeqRecord import SeqRecord
from Bio import BiopythonWarning

from Bio.Phylo import BaseTree


class PhyloXMLWarning(BiopythonWarning):
    """Warning for non-compliance with the phyloXML specification."""

    pass


def _check_str(text, testfunc):
    """Check a string using testfunc, and warn if there's no match (PRIVATE)."""
    if text is not None and not testfunc(text):
        warnings.warn(
            "String %s doesn't match the given regexp" % text,
            PhyloXMLWarning,
            stacklevel=2,
        )


# Core elements


class PhyloElement(BaseTree.TreeElement):
    """Base class for all PhyloXML objects."""


class Phyloxml(PhyloElement):
    """Root node of the PhyloXML document.

    Contains an arbitrary number of Phylogeny elements, possibly followed by
    elements from other namespaces.

    :Parameters:
        attributes : dict
            (XML namespace definitions)
        phylogenies : list
            The phylogenetic trees
        other : list
            Arbitrary non-phyloXML elements, if any

    """

    def __init__(self, attributes, phylogenies=None, other=None):
        """Initialize parameters for PhyloXML object."""
        self.attributes = {
            # standard
            "xmlns:xsi": "http://www.w3.org/2001/XMLSchema-instance",
            "xmlns": "http://www.phyloxml.org",
            "xsi:schemaLocation": "http://www.phyloxml.org http://www.phyloxml.org/1.10/phyloxml.xsd",
        }
        if attributes:
            self.attributes.update(attributes)
        self.phylogenies = phylogenies or []
        self.other = other or []

    def __getitem__(self, index):
        """Get a phylogeny by index or name."""
        if isinstance(index, (int, slice)):
            return self.phylogenies[index]
        if not isinstance(index, str):
            raise KeyError("can't use %s as an index" % type(index))
        for tree in self.phylogenies:
            if tree.name == index:
                return tree
        else:
            raise KeyError("no phylogeny found with name %r" % index)

    def __iter__(self):
        """Iterate through the phylogenetic trees in this object."""
        return iter(self.phylogenies)

    def __len__(self):
        """Return the number of phylogenetic trees in this object."""
        return len(self.phylogenies)

    def __str__(self):
        """Return name of phylogenies in the object."""
        return "%s([%s])" % (
            self.__class__.__name__,
            ",\n".join(map(str, self.phylogenies)),
        )


class Other(PhyloElement):
    """Container for non-phyloXML elements in the tree.

    Usually, an Other object will have either a 'value' or a non-empty list
    of 'children', but not both. This is not enforced here, though.

    :Parameters:
        tag : string
            local tag for the XML node
        namespace : string
            XML namespace for the node -- should not be the default phyloXML
            namespace.
        attributes : dict of strings
            attributes on the XML node
        value : string
            text contained directly within this XML node
        children : list
            child nodes, if any (also ``Other`` instances)

    """

    def __init__(self, tag, namespace=None, attributes=None, value=None, children=None):
        """Initialize values for non-phyloXML elements."""
        self.tag = tag
        self.namespace = namespace
        self.attributes = attributes or {}
        self.value = value
        self.children = children or []

    def __iter__(self):
        """Iterate through the children of this object (if any)."""
        return iter(self.children)


class Phylogeny(PhyloElement, BaseTree.Tree):
    """A phylogenetic tree.

    :Parameters:
        root : Clade
            the root node/clade of this tree
        rooted : bool
            True if this tree is rooted
        rerootable : bool
            True if this tree is rerootable
        branch_length_unit : string
            unit for branch_length values on clades
        name : string
            identifier for this tree, not required to be unique
        id : Id
            unique identifier for this tree
        description : string
            plain-text description
        date : Date
            date for the root node of this tree
        confidences : list
            Confidence objects for this tree
        clade_relations : list
            CladeRelation objects
        sequence_relations : list
            SequenceRelation objects
        properties : list
            Property objects
        other : list
            non-phyloXML elements (type ``Other``)

    """

    def __init__(
        self,
        root=None,
        rooted=True,
        rerootable=None,
        branch_length_unit=None,
        type=None,
        # Child nodes
        name=None,
        id=None,
        description=None,
        date=None,
        # Collections
        confidences=None,
        clade_relations=None,
        sequence_relations=None,
        properties=None,
        other=None,
    ):
        """Initialize values for phylogenetic tree object."""
        assert isinstance(rooted, bool)
        self.root = root
        self.rooted = rooted
        self.rerootable = rerootable
        self.branch_length_unit = branch_length_unit
        self.type = type
        self.name = name
        self.id = id
        self.description = description
        self.date = date
        self.confidences = confidences or []
        self.clade_relations = clade_relations or []
        self.sequence_relations = sequence_relations or []
        self.properties = properties or []
        self.other = other or []

    @classmethod
    def from_tree(cls, tree, **kwargs):
        """Create a new Phylogeny given a Tree (from Newick/Nexus or BaseTree).

        Keyword arguments are the usual ``Phylogeny`` constructor parameters.
        """
        phy = cls(
            root=Clade.from_clade(tree.root),
            rooted=tree.rooted,
            name=tree.name,
            id=(tree.id is not None) and Id(str(tree.id)) or None,
        )
        phy.__dict__.update(kwargs)
        return phy

    @classmethod
    def from_clade(cls, clade, **kwargs):
        """Create a new Phylogeny given a Newick or BaseTree Clade object.

        Keyword arguments are the usual ``PhyloXML.Clade`` constructor parameters.
        """
        return Clade.from_clade(clade).to_phylogeny(**kwargs)

    def as_phyloxml(self):
        """Return this tree, a PhyloXML-compatible Phylogeny object.

        Overrides the ``BaseTree`` method.
        """
        return self

    def to_phyloxml_container(self, **kwargs):
        """Create a new Phyloxml object containing just this phylogeny."""
        return Phyloxml(kwargs, phylogenies=[self])

    def to_alignment(self):
        """Construct an alignment from the aligned sequences in this tree."""

        def is_aligned_seq(elem):
            if isinstance(elem, Sequence) and elem.mol_seq.is_aligned:
                return True
            return False

        seqs = self._filter_search(is_aligned_seq, "preorder", True)
        try:
            first_seq = next(seqs)
        except StopIteration:
            # No aligned sequences were found --> empty MSA
            return MultipleSeqAlignment([])
        msa = MultipleSeqAlignment([first_seq.to_seqrecord()])
        msa.extend(seq.to_seqrecord() for seq in seqs)
        return msa

    # Singular property for plural attribute
    def _get_confidence(self):
        """Equivalent to self.confidences[0] if there is only 1 value (PRIVATE).

        See Also: ``Clade.confidence``, ``Clade.taxonomy``

        """
        if len(self.confidences) == 0:
            return None
        if len(self.confidences) > 1:
            raise AttributeError(
                "more than 1 confidence value available; use Phylogeny.confidences"
            )
        return self.confidences[0]

    def _set_confidence(self, value):
        if value is None:
            # Special case: mirror the behavior of _get_confidence
            self.confidences = []
            return
        if isinstance(value, (float, int)):
            value = Confidence(value)
        elif not isinstance(value, Confidence):
            raise ValueError("value must be a number or Confidence instance")
        if len(self.confidences) == 0:
            self.confidences.append(value)
        elif len(self.confidences) == 1:
            self.confidences[0] = value
        else:
            raise ValueError(
                "multiple confidence values already exist; "
                "use Phylogeny.confidences instead"
            )

    def _del_confidence(self):
        self.confidences = []

    confidence = property(_get_confidence, _set_confidence, _del_confidence)


class Clade(PhyloElement, BaseTree.Clade):
    """Describes a branch of the current phylogenetic tree.

    Used recursively, describes the topology of a phylogenetic tree.

    Both ``color`` and ``width`` elements should be interpreted by client code
    as applying to the whole clade, including all descendents, unless
    overwritten in-sub clades. This module doesn't automatically assign these
    attributes to sub-clades to achieve this cascade -- and neither should you.

    :Parameters:
        branch_length
            parent branch length of this clade
        id_source
            link other elements to a clade (on the xml-level)
        name : string
            short label for this clade
        confidences : list of Confidence objects
            used to indicate the support for a clade/parent branch.
        width : float
            branch width for this clade (including branch from parent)
        color : BranchColor
            color used for graphical display of this clade
        node_id
            unique identifier for the root node of this clade
        taxonomies : list
            Taxonomy objects
        sequences : list
            Sequence objects
        events : Events
            describe such events as gene-duplications at the root node/parent
            branch of this clade
        binary_characters : BinaryCharacters
            binary characters
        distributions : list of Distribution objects
            distribution(s) of this clade
        date : Date
            a date for the root node of this clade
        references : list
            Reference objects
        properties : list
            Property objects
        clades : list Clade objects
            Sub-clades
        other : list of Other objects
            non-phyloXML objects

    """

    def __init__(
        self,
        # Attributes
        branch_length=None,
        id_source=None,
        # Child nodes
        name=None,
        width=None,
        color=None,
        node_id=None,
        events=None,
        binary_characters=None,
        date=None,
        # Collections
        confidences=None,
        taxonomies=None,
        sequences=None,
        distributions=None,
        references=None,
        properties=None,
        clades=None,
        other=None,
    ):
        """Initialize value for the Clade object."""
        self.branch_length = branch_length
        self.id_source = id_source
        self.name = name
        self.width = width
        self.color = color
        self.node_id = node_id
        self.events = events
        self.binary_characters = binary_characters
        self.date = date
        self.confidences = confidences or []
        self.taxonomies = taxonomies or []
        self.sequences = sequences or []
        self.distributions = distributions or []
        self.references = references or []
        self.properties = properties or []
        self.clades = clades or []
        self.other = other or []

    @classmethod
    def from_clade(cls, clade, **kwargs):
        """Create a new PhyloXML Clade from a Newick or BaseTree Clade object.

        Keyword arguments are the usual PhyloXML Clade constructor parameters.
        """
        new_clade = cls(branch_length=clade.branch_length, name=clade.name)
        new_clade.clades = [cls.from_clade(c) for c in clade]
        new_clade.confidence = clade.confidence
        new_clade.width = clade.width
        new_clade.color = (
            BranchColor(clade.color.red, clade.color.green, clade.color.blue)
            if clade.color
            else None
        )
        new_clade.__dict__.update(kwargs)
        return new_clade

    def to_phylogeny(self, **kwargs):
        """Create a new phylogeny containing just this clade."""
        phy = Phylogeny(root=self, date=self.date)
        phy.__dict__.update(kwargs)
        return phy

    # Shortcuts for list attributes that are usually only 1 item
    # NB: Duplicated from Phylogeny class
    def _get_confidence(self):
        """Return confidence values (PRIVATE)."""
        if len(self.confidences) == 0:
            return None
        if len(self.confidences) > 1:
            raise AttributeError(
                "more than 1 confidence value available; use Clade.confidences"
            )
        return self.confidences[0]

    def _set_confidence(self, value):
        """Set the confidence value (PRIVATE)."""
        if value is None:
            # Special case: mirror the behavior of _get_confidence
            self.confidences = []
            return
        if isinstance(value, (float, int)):
            value = Confidence(value)
        elif not isinstance(value, Confidence):
            raise ValueError("value must be a number or Confidence instance")
        if len(self.confidences) == 0:
            self.confidences.append(value)
        elif len(self.confidences) == 1:
            self.confidences[0] = value
        else:
            raise ValueError(
                "multiple confidence values already exist; "
                "use Phylogeny.confidences instead"
            )

    def _del_confidence(self):
        """Delete confidences values (PRIVATE)."""
        self.confidences = []

    confidence = property(_get_confidence, _set_confidence, _del_confidence)

    def _get_taxonomy(self):
        """Get taxonomy list for the clade (PRIVATE)."""
        if len(self.taxonomies) == 0:
            return None
        if len(self.taxonomies) > 1:
            raise AttributeError(
                "more than 1 taxonomy value available; use Clade.taxonomies"
            )
        return self.taxonomies[0]

    def _set_taxonomy(self, value):
        """Set a taxonomy for the clade (PRIVATE)."""
        if not isinstance(value, Taxonomy):
            raise ValueError("assigned value must be a Taxonomy instance")
        if len(self.taxonomies) == 0:
            self.taxonomies.append(value)
        elif len(self.taxonomies) == 1:
            self.taxonomies[0] = value
        else:
            raise ValueError(
                "multiple taxonomy values already exist; "
                "use Phylogeny.taxonomies instead"
            )

    taxonomy = property(_get_taxonomy, _set_taxonomy)


# PhyloXML wrapper for a special BaseTree attribute


class BranchColor(PhyloElement, BaseTree.BranchColor):
    """Manage Tree branch's color."""

    def __init__(self, *args, **kwargs):
        """Initialize parameters for the BranchColor object."""
        BaseTree.BranchColor.__init__(self, *args, **kwargs)


# PhyloXML-specific complex types


class Accession(PhyloElement):
    """Captures the local part in a sequence identifier.

    Example: In ``UniProtKB:P17304``, the Accession instance attribute ``value``
    is 'P17304' and the ``source`` attribute is 'UniProtKB'.
    """

    def __init__(self, value, source):
        """Initialize value for Accession object."""
        self.value = value
        self.source = source

    def __str__(self):
        """Show the class name and an identifying attribute."""
        return "%s:%s" % (self.source, self.value)


class Annotation(PhyloElement):
    """The annotation of a molecular sequence.

    It is recommended to annotate by using the optional 'ref' attribute.

    :Parameters:
        ref : string
            reference string, e.g. 'GO:0008270',
            'KEGG:Tetrachloroethene degradation', 'EC:1.1.1.1'
        source : string
            plain-text source for this annotation
        evidence : str
            describe evidence as free text (e.g. 'experimental')
        desc : string
            free text description
        confidence : Confidence
            state the type and value of support (type Confidence)
        properties : list
            typed and referenced annotations from external resources
        uri : Uri
            link

    """

    re_ref = re.compile(r"[a-zA-Z0-9_]+:[a-zA-Z0-9_\.\-\s]+")

    def __init__(
        self,
        # Attributes
        ref=None,
        source=None,
        evidence=None,
        type=None,
        # Child nodes
        desc=None,
        confidence=None,
        uri=None,
        # Collection
        properties=None,
    ):
        """Initialize value for the Annotation object."""
        _check_str(ref, self.re_ref.match)
        self.ref = ref
        self.source = source
        self.evidence = evidence
        self.type = type
        self.desc = desc
        self.confidence = confidence
        self.uri = uri
        self.properties = properties or []


class BinaryCharacters(PhyloElement):
    """Binary characters at the root of a clade.

    The names and/or counts of binary characters present, gained, and lost
    at the root of a clade.
    """

    def __init__(
        self,
        # Attributes
        type=None,
        gained_count=None,
        lost_count=None,
        present_count=None,
        absent_count=None,
        # Child nodes (flattened into collections)
        gained=None,
        lost=None,
        present=None,
        absent=None,
    ):
        """Initialize values for the BinaryCharacters object."""
        self.type = type
        self.gained_count = gained_count
        self.lost_count = lost_count
        self.present_count = present_count
        self.absent_count = absent_count
        self.gained = gained or []
        self.lost = lost or []
        self.present = present or []
        self.absent = absent or []


class CladeRelation(PhyloElement):
    """Expresses a typed relationship between two clades.

    For example, this could be used to describe multiple parents of a clade.

    :type id_ref_0: str
    :type id_ref_1: str
    :type distance: str
    :type type: str

    :type confidence: Confidence
    """

    def __init__(self, type, id_ref_0, id_ref_1, distance=None, confidence=None):
        """Initialize values for the CladeRelation object."""
        self.distance = distance
        self.type = type
        self.id_ref_0 = id_ref_0
        self.id_ref_1 = id_ref_1
        self.confidence = confidence


class Confidence(PhyloElement):
    """A general purpose confidence element.

    For example, this can be used to express the bootstrap support value of a
    clade (in which case the ``type`` attribute is 'bootstrap').

    :Parameters:
        value : float
            confidence value
        type : string
            label for the type of confidence, e.g. 'bootstrap'

    """

    def __init__(self, value, type="unknown"):
        """Initialize values for the Confidence object."""
        self.value = value
        self.type = type

    # Comparison operators

    def __hash__(self):
        """Return the hash value of the object.

        Hash values are integers. They are used to quickly compare dictionary
        keys during a dictionary lookup. Numeric values that compare equal have
        the same hash value (even if they are of different types, as is the
        case for 1 and 1.0).
        """
        return id(self)

    def __eq__(self, other):
        """Check for equality between Confidence objects."""
        if isinstance(other, Confidence):
            return self.value == other.value
        return self.value == other

    def __ne__(self, other):
        """Check for inequality between two Confidence objects."""
        if isinstance(other, Confidence):
            return self.value != other.value
        return self.value != other

    # Ordering -- see functools.total_ordering in Py2.7

    def __lt__(self, other):
        """Return True if confidence is less than the other confidence."""
        if isinstance(other, Confidence):
            return self.value < other.value
        return self.value < other

    def __le__(self, other):
        """Return True if confidence is less than or equal to the other confidence."""
        return self < other or self == other

    def __gt__(self, other):
        """Return True if confidence is grater than the other confidence."""
        return not (self <= other)

    def __ge__(self, other):
        """Return True if confidence is grater than or equal to the other confidence."""
        return not (self.value < other)

    # Arithmetic operators, including reverse

    def __add__(self, other):
        """Conduct addition between values of two Confidence objects."""
        return self.value + other

    def __radd__(self, other):
        """Conduct reverse addition between values of two Confidence objects."""
        return other + self.value

    def __sub__(self, other):
        """Conduct subtraction between values of two Confidence objects."""
        return self.value - other

    def __rsub__(self, other):
        """Conduct reverse subtraction between values of two Confidence objects."""
        return other - self.value

    def __mul__(self, other):
        """Conduct multiplication between values of two Confidence objects."""
        return self.value * other

    def __rmul__(self, other):
        """Conduct reverse multiplication between values of two Confidence objects."""
        return other * self.value

    def __truediv__(self, other):
        """Conduct division between values of two Confidence objects."""
        return self.value.__truediv__(other)

    def __rtruediv__(self, other):
        """Conduct reverse division between values of two Confidence objects."""
        return other.__rtruediv__(self.value)

    def __floordiv__(self, other):
        """C-style and old-style division."""
        return self.value.__floordiv__(other)

    def __rfloordiv__(self, other):
        """Conduct revers C-style and old-style division."""
        return other.__floordiv__(self.value)

    def __mod__(self, other):
        """Conduct modulus between values of two Confidence objects."""
        return self.value % other

    def __rmod__(self, other):
        """Conduct reverse modulus between values of two Confidence objects."""
        return other % self.value

    def __divmod__(self, other):
        """Return quotient and remainder, dividing the Confidence value by the given value."""
        return divmod(self.value, other)

    def __rdivmod__(self, other):
        """Return quotient and remainder, dividing the given value by the Confidence value."""
        return divmod(other, self.value)

    def __pow__(self, other, modulo=None):
        """Return the value of Confidence object raised to the given power."""
        if modulo is not None:
            return pow(self.value, other, modulo)
        return pow(self.value, other)

    def __rpow__(self, other):
        """Return the given value raised to the power of the Confidence object value."""
        return pow(other, self.value)

    # Unary arithmetic operations: -, +, abs()

    def __neg__(self):
        """Conduct negation of a Confidence object."""
        return -self.value

    def __pos__(self):
        """Return the value of a Confidence object."""
        return self.value

    def __abs__(self):
        """Return absolute value of Confidence object."""
        return abs(self.value)

    # Explicit coercion to numeric types: float, int

    def __float__(self):
        """Return float value of Confidence object."""
        return float(self.value)

    def __int__(self):
        """Return integer value of Confidence object."""
        return int(self.value)


class Date(PhyloElement):
    """A date associated with a clade/node.

    Its value can be numerical by using the 'value' element and/or free text
    with the 'desc' element' (e.g. 'Silurian'). If a numerical value is used, it
    is recommended to employ the 'unit' attribute.

    :Parameters:
        unit : string
            type of numerical value (e.g. 'mya' for 'million years ago')
        value : float
            the date value
        desc : string
            plain-text description of the date
        minimum : float
            lower bound on the date value
        maximum : float
            upper bound on the date value

    """

    def __init__(self, value=None, unit=None, desc=None, minimum=None, maximum=None):
        """Initialize values of the Date object."""
        self.value = value
        self.unit = unit
        self.desc = desc
        self.minimum = minimum
        self.maximum = maximum

    def __str__(self):
        """Show the class name and the human-readable date."""
        if self.unit and self.value is not None:
            return "%s %s" % (self.value, self.unit)
        if self.desc is not None:
            return self.desc
        return self.__class__.__name__


class Distribution(PhyloElement):
    """Geographic distribution of the items of a clade (species, sequences).

    Intended for phylogeographic applications.

    :Parameters:
        desc : string
            free-text description of the location
        points : list of ``Point`` objects
            coordinates (similar to the 'Point' element in Google's KML format)
        polygons : list of ``Polygon`` objects
            coordinate sets defining geographic regions

    """

    def __init__(self, desc=None, points=None, polygons=None):
        """Initialize values of Distribution object."""
        self.desc = desc
        self.points = points or []
        self.polygons = polygons or []


class DomainArchitecture(PhyloElement):
    """Domain architecture of a protein.

    :Parameters:
        length : int
            total length of the protein sequence
        domains : list ProteinDomain objects
            the domains within this protein

    """

    def __init__(self, length=None, domains=None):
        """Initialize values of the DomainArchitecture object."""
        self.length = length
        self.domains = domains


class Events(PhyloElement):
    """Events at the root node of a clade (e.g. one gene duplication).

    All attributes are set to None by default, but this object can also be
    treated as a dictionary, in which case None values are treated as missing
    keys and deleting a key resets that attribute's value back to None.
    """

    ok_type = {
        "transfer",
        "fusion",
        "speciation_or_duplication",
        "other",
        "mixed",
        "unassigned",
    }

    def __init__(
        self,
        type=None,
        duplications=None,
        speciations=None,
        losses=None,
        confidence=None,
    ):
        """Initialize values of the Events object."""
        _check_str(type, self.ok_type.__contains__)
        self.type = type
        self.duplications = duplications
        self.speciations = speciations
        self.losses = losses
        self.confidence = confidence

    def items(self):
        """Return Event's items."""
        return [(k, v) for k, v in self.__dict__.items() if v is not None]

    def keys(self):
        """Return Event's keys."""
        return [k for k, v in self.__dict__.items() if v is not None]

    def values(self):
        """Return values from a key-value pair in an Events dict."""
        return [v for v in self.__dict__.values() if v is not None]

    def __len__(self):
        """Return number of Events."""
        # TODO - Better way to do this?
        return len(self.values())

    def __getitem__(self, key):
        """Get value of Event with the given key."""
        try:
            val = getattr(self, key)
        except AttributeError:
            raise KeyError(key) from None
        if val is None:
            raise KeyError("%r has not been set in this object" % key)
        return val

    def __setitem__(self, key, val):
        """Add item to Event dict."""
        setattr(self, key, val)

    def __delitem__(self, key):
        """Delete Event with given key."""
        setattr(self, key, None)

    def __iter__(self):
        """Iterate over the keys present in a Events dict."""
        return iter(self.keys())

    def __contains__(self, key):
        """Return True if Event dict contains key."""
        try:
            return getattr(self, key) is not None
        except AttributeError:
            return False


class Id(PhyloElement):
    """A general-purpose identifier element.

    Allows to indicate the provider (or authority) of an identifier, e.g. NCBI,
    along with the value itself.
    """

    def __init__(self, value, provider=None):
        """Initialize values for the identifier object."""
        self.value = value
        self.provider = provider

    def __str__(self):
        """Return identifier as a string."""
        if self.provider is not None:
            return "%s:%s" % (self.provider, self.value)
        return self.value


class MolSeq(PhyloElement):
    """Store a molecular sequence.

    :Parameters:
        value : string
            the sequence itself
        is_aligned : bool
            True if this sequence is aligned with the others (usually meaning
            all aligned seqs are the same length and gaps may be present)

    """

    re_value = re.compile(r"[a-zA-Z\.\-\?\*_]+")

    def __init__(self, value, is_aligned=None):
        """Initialize parameters for the MolSeq object."""
        _check_str(value, self.re_value.match)
        self.value = value
        self.is_aligned = is_aligned

    def __str__(self):
        """Return the value of the Molecular Sequence object."""
        return self.value


class Point(PhyloElement):
    """Geographic coordinates of a point, with an optional altitude.

    Used by element 'Distribution'.

    :Parameters:
        geodetic_datum : string, required
            the geodetic datum (also called 'map datum'). For example, Google's
            KML uses 'WGS84'.
        lat : numeric
            latitude
        long : numeric
            longitude
        alt : numeric
            altitude
        alt_unit : string
            unit for the altitude (e.g. 'meter')

    """

    def __init__(self, geodetic_datum, lat, long, alt=None, alt_unit=None):
        """Initialize value for the Point object."""
        self.geodetic_datum = geodetic_datum
        self.lat = lat
        self.long = long
        self.alt = alt
        self.alt_unit = alt_unit


class Polygon(PhyloElement):
    """A polygon defined by a list of 'Points' (used by element 'Distribution').

    :param points: list of 3 or more points representing vertices.

    """

    def __init__(self, points=None):
        """Initialize value for the Polygon object."""
        self.points = points or []

    def __str__(self):
        """Return list of points as a string."""
        return "%s([%s])" % (self.__class__.__name__, ",\n".join(map(str, self.points)))


class Property(PhyloElement):
    """A typed and referenced property from an external resources.

    Can be attached to ``Phylogeny``, ``Clade``, and ``Annotation`` objects.

    :Parameters:
        value : string
            the value of the property
        ref : string
            reference to an external resource, e.g. "NOAA:depth"
        applies_to : string
            indicates the item to which a property applies to (e.g.  'node' for
            the parent node of a clade, 'parent_branch' for the parent branch of
            a clade, or just 'clade').
        datatype : string
            the type of a property; limited to xsd-datatypes
            (e.g. 'xsd:string', 'xsd:boolean', 'xsd:integer', 'xsd:decimal',
            'xsd:float', 'xsd:double', 'xsd:date', 'xsd:anyURI').
        unit : string (optional)
            the unit of the property, e.g. "METRIC:m"
        id_ref : Id (optional)
            allows to attached a property specifically to one element (on the
            xml-level)

    """

    re_ref = re.compile(r"[a-zA-Z0-9_]+:[a-zA-Z0-9_\.\-\s]+")
    ok_applies_to = {
        "phylogeny",
        "clade",
        "node",
        "annotation",
        "parent_branch",
        "other",
    }
    ok_datatype = {
        "xsd:string",
        "xsd:boolean",
        "xsd:decimal",
        "xsd:float",
        "xsd:double",
        "xsd:duration",
        "xsd:dateTime",
        "xsd:time",
        "xsd:date",
        "xsd:gYearMonth",
        "xsd:gYear",
        "xsd:gMonthDay",
        "xsd:gDay",
        "xsd:gMonth",
        "xsd:hexBinary",
        "xsd:base64Binary",
        "xsd:anyURI",
        "xsd:normalizedString",
        "xsd:token",
        "xsd:integer",
        "xsd:nonPositiveInteger",
        "xsd:negativeInteger",
        "xsd:long",
        "xsd:int",
        "xsd:short",
        "xsd:byte",
        "xsd:nonNegativeInteger",
        "xsd:unsignedLong",
        "xsd:unsignedInt",
        "xsd:unsignedShort",
        "xsd:unsignedByte",
        "xsd:positiveInteger",
    }

    def __init__(self, value, ref, applies_to, datatype, unit=None, id_ref=None):
        """Initialize value for the Property object."""
        _check_str(ref, self.re_ref.match)
        _check_str(applies_to, self.ok_applies_to.__contains__)
        _check_str(datatype, self.ok_datatype.__contains__)
        _check_str(unit, self.re_ref.match)
        self.unit = unit
        self.id_ref = id_ref
        self.value = value
        self.ref = ref
        self.applies_to = applies_to
        self.datatype = datatype


class ProteinDomain(PhyloElement):
    """Represents an individual domain in a domain architecture.

    The locations use 0-based indexing, as most Python objects including
    SeqFeature do, rather than the usual biological convention starting at 1.
    This means the start and end attributes can be used directly as slice
    indexes on Seq objects.

    :Parameters:
        start : non-negative integer
            start of the domain on the sequence, using 0-based indexing
        end : non-negative integer
            end of the domain on the sequence
        confidence : float
            can be used to store e.g. E-values
        id : string
            unique identifier/name

    """

    def __init__(self, value, start, end, confidence=None, id=None):
        """Initialize value for a ProteinDomain object."""
        self.value = value
        self.start = start
        self.end = end
        self.confidence = confidence
        self.id = id

    @classmethod
    def from_seqfeature(cls, feat):
        """Create ProteinDomain object from SeqFeature."""
        return ProteinDomain(
            feat.id,
            feat.location.nofuzzy_start,
            feat.location.nofuzzy_end,
            confidence=feat.qualifiers.get("confidence"),
        )

    def to_seqfeature(self):
        """Create a SeqFeature from the ProteinDomain Object."""
        feat = SeqFeature(location=FeatureLocation(self.start, self.end), id=self.value)
        try:
            confidence = self.confidence
        except AttributeError:
            pass
        else:
            feat.qualifiers["confidence"] = confidence
        return feat


class Reference(PhyloElement):
    """Literature reference for a clade.

    NB: Whenever possible, use the ``doi`` attribute instead of the free-text
    ``desc`` element.
    """

    re_doi = re.compile(r"[a-zA-Z0-9_\.]+/[a-zA-Z0-9_\.]+")

    def __init__(self, doi=None, desc=None):
        """Initialize elements of the Reference class object."""
        _check_str(doi, self.re_doi.match)
        self.doi = doi
        self.desc = desc


class Sequence(PhyloElement):
    """A molecular sequence (Protein, DNA, RNA) associated with a node.

    One intended use for ``id_ref`` is to link a sequence to a taxonomy (via the
    taxonomy's ``id_source``) in case of multiple sequences and taxonomies per
    node.

    :Parameters:
        type : {'dna', 'rna', 'protein'}
            type of molecule this sequence represents
        id_ref : string
            reference to another resource
        id_source : string
            source for the reference
        symbol : string
            short symbol of the sequence, e.g. 'ACTM' (max. 10 chars)
        accession : Accession
            accession code for this sequence.
        name : string
            full name of the sequence, e.g. 'muscle Actin'
        location
            location of a sequence on a genome/chromosome.
        mol_seq : MolSeq
            the molecular sequence itself
        uri : Uri
            link
        annotations : list of Annotation objects
            annotations on this sequence
        domain_architecture : DomainArchitecture
            protein domains on this sequence
        other : list of Other objects
            non-phyloXML elements

    """

    types = {"dna", "rna", "protein"}
    re_symbol = re.compile(r"\S{1,10}")

    def __init__(
        self,
        # Attributes
        type=None,
        id_ref=None,
        id_source=None,
        # Child nodes
        symbol=None,
        accession=None,
        name=None,
        location=None,
        mol_seq=None,
        uri=None,
        domain_architecture=None,
        # Collections
        annotations=None,
        other=None,
    ):
        """Initialize value for a Sequence object."""
        _check_str(type, self.types.__contains__)
        _check_str(symbol, self.re_symbol.match)
        self.type = type
        self.id_ref = id_ref
        self.id_source = id_source
        self.symbol = symbol
        self.accession = accession
        self.name = name
        self.location = location
        self.mol_seq = mol_seq
        self.uri = uri
        self.domain_architecture = domain_architecture
        self.annotations = annotations or []
        self.other = other or []

    @classmethod
    def from_seqrecord(cls, record, is_aligned=None):
        """Create a new PhyloXML Sequence from a SeqRecord object."""
        if is_aligned is None:
            is_aligned = "-" in record.seq
        params = {
            "accession": Accession(record.id, ""),
            "symbol": record.name,
            "name": record.description,
            "mol_seq": MolSeq(str(record.seq), is_aligned),
        }
        molecule_type = record.annotations.get("molecule_type")
        if molecule_type is not None:
            if "DNA" in molecule_type:
                params["type"] = "dna"
            elif "RNA" in molecule_type:
                params["type"] = "rna"
            elif "protein" in molecule_type:
                params["type"] = "protein"

        # Unpack record.annotations
        for key in ("id_ref", "id_source", "location"):
            if key in record.annotations:
                params[key] = record.annotations[key]
        if isinstance(record.annotations.get("uri"), dict):
            params["uri"] = Uri(**record.annotations["uri"])
        # Build a Sequence.annotation object
        if record.annotations.get("annotations"):
            params["annotations"] = []
            for annot in record.annotations["annotations"]:
                ann_args = {}
                for key in ("ref", "source", "evidence", "type", "desc"):
                    if key in annot:
                        ann_args[key] = annot[key]
                if isinstance(annot.get("confidence"), list):
                    ann_args["confidence"] = Confidence(*annot["confidence"])
                if isinstance(annot.get("properties"), list):
                    ann_args["properties"] = [
                        Property(**prop)
                        for prop in annot["properties"]
                        if isinstance(prop, dict)
                    ]
                params["annotations"].append(Annotation(**ann_args))

        # Unpack record.features
        if record.features:
            params["domain_architecture"] = DomainArchitecture(
                length=len(record.seq),
                domains=[
                    ProteinDomain.from_seqfeature(feat) for feat in record.features
                ],
            )

        return Sequence(**params)

    def to_seqrecord(self):
        """Create a SeqRecord object from this Sequence instance.

        The seqrecord.annotations dictionary is packed like so::

            { # Sequence attributes with no SeqRecord equivalent:
              'id_ref': self.id_ref,
              'id_source': self.id_source,
              'location': self.location,
              'uri': { 'value': self.uri.value,
                              'desc': self.uri.desc,
                              'type': self.uri.type },
              # Sequence.annotations attribute (list of Annotations)
              'annotations': [{'ref': ann.ref,
                               'source': ann.source,
                               'evidence': ann.evidence,
                               'type': ann.type,
                               'confidence': [ann.confidence.value,
                                              ann.confidence.type],
                               'properties': [{'value': prop.value,
                                                'ref': prop.ref,
                                                'applies_to': prop.applies_to,
                                                'datatype': prop.datatype,
                                                'unit': prop.unit,
                                                'id_ref': prop.id_ref}
                                               for prop in ann.properties],
                              } for ann in self.annotations],
            }

        """

        def clean_dict(dct):
            """Remove None-valued items from a dictionary."""
            return {key: val for key, val in dct.items() if val is not None}

        seqrec = SeqRecord(
            Seq(self.mol_seq.value),
            **clean_dict(
                {
                    "id": str(self.accession),
                    "name": self.symbol,
                    "description": self.name,
                    # 'dbxrefs': None,
                }
            )
        )
        if self.domain_architecture:
            seqrec.features = [
                dom.to_seqfeature() for dom in self.domain_architecture.domains
            ]
        # Sequence attributes with no SeqRecord equivalent
        if self.type == "dna":
            molecule_type = "DNA"
        elif self.type == "rna":
            molecule_type = "RNA"
        elif self.type == "protein":
            molecule_type = "protein"
        else:
            molecule_type = None
        seqrec.annotations = clean_dict(
            {
                "id_ref": self.id_ref,
                "id_source": self.id_source,
                "location": self.location,
                "uri": self.uri
                and clean_dict(
                    {
                        "value": self.uri.value,
                        "desc": self.uri.desc,
                        "type": self.uri.type,
                    }
                ),
                "molecule_type": molecule_type,
                "annotations": self.annotations
                and [
                    clean_dict(
                        {
                            "ref": ann.ref,
                            "source": ann.source,
                            "evidence": ann.evidence,
                            "type": ann.type,
                            "confidence": ann.confidence
                            and [ann.confidence.value, ann.confidence.type],
                            "properties": [
                                clean_dict(
                                    {
                                        "value": prop.value,
                                        "ref": prop.ref,
                                        "applies_to": prop.applies_to,
                                        "datatype": prop.datatype,
                                        "unit": prop.unit,
                                        "id_ref": prop.id_ref,
                                    }
                                )
                                for prop in ann.properties
                            ],
                        }
                    )
                    for ann in self.annotations
                ],
            }
        )
        return seqrec


class SequenceRelation(PhyloElement):
    """Express a typed relationship between two sequences.

    For example, this could be used to describe an orthology (in which case
    attribute 'type' is 'orthology').

    :Parameters:
        id_ref_0 : Id
            first sequence reference identifier
        id_ref_1 : Id
            second sequence reference identifier
        distance : float
            distance between the two sequences
        type : restricted string
            describe the type of relationship
        confidence : Confidence
            confidence value for this relation

    """

    ok_type = {
        "orthology",
        "one_to_one_orthology",
        "super_orthology",
        "paralogy",
        "ultra_paralogy",
        "xenology",
        "unknown",
        "other",
    }

    def __init__(self, type, id_ref_0, id_ref_1, distance=None, confidence=None):
        """Initialize the class."""
        _check_str(type, self.ok_type.__contains__)
        self.distance = distance
        self.type = type
        self.id_ref_0 = id_ref_0
        self.id_ref_1 = id_ref_1
        self.confidence = confidence


class Taxonomy(PhyloElement):
    """Describe taxonomic information for a clade.

    :Parameters:
        id_source : Id
            link other elements to a taxonomy (on the XML level)
        id : Id
            unique identifier of a taxon, e.g. Id('6500',
            provider='ncbi_taxonomy') for the California sea hare
        code : restricted string
            store UniProt/Swiss-Prot style organism codes, e.g. 'APLCA' for the
            California sea hare 'Aplysia californica'
        scientific_name : string
            the standard scientific name for this organism, e.g. 'Aplysia
            californica' for the California sea hare
        authority : string
            keep the authority, such as 'J. G. Cooper, 1863', associated with
            the 'scientific_name'
        common_names : list of strings
            common names for this organism
        synonyms : list of strings
            synonyms for this taxon?
        rank : restricted string
            taxonomic rank
        uri : Uri
            link
        other : list of Other objects
            non-phyloXML elements

    """

    re_code = re.compile(r"[a-zA-Z0-9_]{2,10}")
    ok_rank = {
        "domain",
        "kingdom",
        "subkingdom",
        "branch",
        "infrakingdom",
        "superphylum",
        "phylum",
        "subphylum",
        "infraphylum",
        "microphylum",
        "superdivision",
        "division",
        "subdivision",
        "infradivision",
        "superclass",
        "class",
        "subclass",
        "infraclass",
        "superlegion",
        "legion",
        "sublegion",
        "infralegion",
        "supercohort",
        "cohort",
        "subcohort",
        "infracohort",
        "superorder",
        "order",
        "suborder",
        "superfamily",
        "family",
        "subfamily",
        "supertribe",
        "tribe",
        "subtribe",
        "infratribe",
        "genus",
        "subgenus",
        "superspecies",
        "species",
        "subspecies",
        "variety",
        "subvariety",
        "form",
        "subform",
        "cultivar",
        "unknown",
        "other",
    }

    def __init__(
        self,
        # Attributes
        id_source=None,
        # Child nodes
        id=None,
        code=None,
        scientific_name=None,
        authority=None,
        rank=None,
        uri=None,
        # Collections
        common_names=None,
        synonyms=None,
        other=None,
    ):
        """Initialize the class."""
        _check_str(code, self.re_code.match)
        _check_str(rank, self.ok_rank.__contains__)
        self.id_source = id_source
        self.id = id
        self.code = code
        self.scientific_name = scientific_name
        self.authority = authority
        self.rank = rank
        self.uri = uri
        self.common_names = common_names or []
        self.synonyms = synonyms or []
        self.other = other or []

    def __str__(self):
        """Show the class name and an identifying attribute."""
        if self.code is not None:
            return self.code
        if self.scientific_name is not None:
            return self.scientific_name
        if self.rank is not None:
            return self.rank
        if self.id is not None:
            return str(self.id)
        return self.__class__.__name__


class Uri(PhyloElement):
    """A uniform resource identifier.

    In general, this is expected to be an URL (for example, to link to an image
    on a website, in which case the ``type`` attribute might be 'image' and
    ``desc`` might be 'image of a California sea hare').
    """

    def __init__(self, value, desc=None, type=None):
        """Initialize the class."""
        self.value = value
        self.desc = desc
        self.type = type

    def __str__(self):
        """Return string representation of Uri."""
        if self.value:
            return self.value
        return repr(self)



# Copyright 2019 by Michiel de Hoon.
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Substitution matrices."""

import os
import string
import numpy


class Array(numpy.ndarray):
    """numpy array subclass indexed by integers and by letters."""

    def __new__(cls, alphabet=None, dims=None, data=None, dtype=float):
        """Create a new Array instance."""
        if isinstance(data, dict):
            if alphabet is not None:
                raise ValueError("alphabet should be None if data is a dict")
            if dims is not None:
                raise ValueError("dims should be None if data is a dict")
            alphabet = []
            for key in data:
                if isinstance(key, str):
                    if dims is None:
                        dims = 1
                    elif dims != 1:
                        raise ValueError("inconsistent dimensions in data")
                    alphabet.append(key)
                elif isinstance(key, tuple):
                    single_letters = True
                    if dims is None:
                        dims = len(key)
                    elif dims != len(key):
                        raise ValueError("inconsistent dimensions in data")
                    if dims == 1:
                        if not isinstance(key, str):
                            raise ValueError("expected string")
                        if len(key) > 1:
                            single_letters = False
                        alphabet.append(key)
                    elif dims == 2:
                        for letter in key:
                            if not isinstance(letter, str):
                                raise ValueError("expected string")
                            if len(letter) > 1:
                                single_letters = False
                            alphabet.append(letter)
                    else:
                        raise ValueError(
                            "data array should be 1- or 2- dimensional "
                            "(found %d dimensions) in key" % dims
                        )
            alphabet = sorted(set(alphabet))
            if single_letters:
                alphabet = "".join(alphabet)
            else:
                alphabet = tuple(alphabet)
            n = len(alphabet)
            if dims == 1:
                shape = (n,)
            elif dims == 2:
                shape = (n, n)
            else:  # dims is None
                raise ValueError("data is an empty dictionary")
            obj = super().__new__(cls, shape, dtype)
            if dims == 1:
                for i, key in enumerate(alphabet):
                    obj[i] = data.get(letter, 0.0)
            elif dims == 2:
                for i1, letter1 in enumerate(alphabet):
                    for i2, letter2 in enumerate(alphabet):
                        key = (letter1, letter2)
                        value = data.get(key, 0.0)
                        obj[i1, i2] = value
            obj._alphabet = alphabet
            return obj
        if alphabet is None:
            alphabet = string.ascii_uppercase
        elif not (isinstance(alphabet, str) or isinstance(alphabet, tuple)):
            raise ValueError("alphabet should be a string or a tuple")
        n = len(alphabet)
        if data is None:
            if dims is None:
                dims = 1
            elif dims not in (1, 2):
                raise ValueError("dims should be 1 or 2 (found %s)" % str(dims))
            shape = (n,) * dims
        else:
            if dims is None:
                shape = data.shape
                dims = len(shape)
                if dims == 1:
                    pass
                elif dims == 2:
                    if shape[0] != shape[1]:
                        raise ValueError("data array is not square")
                else:
                    raise ValueError(
                        "data array should be 1- or 2- dimensional "
                        "(found %d dimensions) " % dims
                    )
            else:
                shape = (n,) * dims
                if data.shape != shape:
                    raise ValueError(
                        "data shape has inconsistent shape (expected (%s), found (%s))"
                        % (shape, data.shape)
                    )
        obj = super().__new__(cls, shape, dtype)
        if data is None:
            obj[:] = 0.0
        else:
            obj[:] = data
        obj._alphabet = alphabet
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self._alphabet = getattr(obj, "_alphabet", None)

    def _convert_key(self, key):
        if isinstance(key, tuple):
            indices = []
            for index in key:
                if isinstance(index, str):
                    try:
                        index = self._alphabet.index(index)
                    except ValueError:
                        raise IndexError("'%s'" % index) from None
                indices.append(index)
            key = tuple(indices)
        elif isinstance(key, str):
            try:
                key = self._alphabet.index(key)
            except ValueError:
                raise IndexError("'%s'" % key) from None
        return key

    def __getitem__(self, key):
        key = self._convert_key(key)
        value = numpy.ndarray.__getitem__(self, key)
        if value.ndim == 2:
            if self.ndim == 2:
                if value.shape != self.shape:
                    raise IndexError("Requesting truncated array")
            elif self.ndim == 1:
                length = self.shape[0]
                if value.shape[0] == length and value.shape[1] == 1:
                    pass
                elif value.shape[0] == 1 and value.shape[1] == length:
                    pass
                else:
                    raise IndexError("Requesting truncated array")
        elif value.ndim == 1:
            if value.shape[0] != self.shape[0]:
                value._alphabet = self.alphabet[key]
        return value.view(Array)

    def __setitem__(self, key, value):
        key = self._convert_key(key)
        numpy.ndarray.__setitem__(self, key, value)

    def __contains__(self, key):
        # Follow dict definition of __contains__
        return key in self.keys()

    def __array_prepare__(self, out_arr, context=None):
        # needed for numpy older than 1.13.0
        ufunc, inputs, i = context
        alphabet = self.alphabet
        for arg in inputs:
            if isinstance(arg, Array):
                if arg.alphabet != alphabet:
                    raise ValueError("alphabets are inconsistent")
        return numpy.ndarray.__array_prepare__(self, out_arr, context)

    def __array_wrap__(self, out_arr, context=None):
        if len(out_arr) == 1:
            return out_arr[0]
        return numpy.ndarray.__array_wrap__(self, out_arr, context)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        args = []
        alphabet = self._alphabet
        for arg in inputs:
            if isinstance(arg, Array):
                if arg.alphabet != alphabet:
                    raise ValueError("alphabets are inconsistent")
                args.append(arg.view(numpy.ndarray))
            else:
                args.append(arg)

        outputs = kwargs.pop("out", None)
        if outputs:
            out_args = []
            for arg in outputs:
                if isinstance(arg, Array):
                    if arg.alphabet != alphabet:
                        raise ValueError("alphabets are inconsistent")
                    out_args.append(arg.view(numpy.ndarray))
                else:
                    out_args.append(arg)
            kwargs["out"] = tuple(out_args)
        else:
            outputs = (None,) * ufunc.nout

        raw_results = super().__array_ufunc__(ufunc, method, *args, **kwargs)
        if raw_results is NotImplemented:
            return NotImplemented

        if method == "at":
            return

        if ufunc.nout == 1:
            raw_results = (raw_results,)

        results = []
        for raw_result, output in zip(raw_results, outputs):
            if raw_result.ndim == 0:
                result = raw_result
            elif output is None:
                result = numpy.asarray(raw_result).view(Array)
                result._alphabet = self._alphabet
            else:
                result = output
                result._alphabet = self._alphabet
            results.append(result)

        return results[0] if len(results) == 1 else results

    def transpose(self, axes=None):
        """Transpose the array."""
        other = numpy.ndarray.transpose(self, axes)
        other._alphabet = self._alphabet
        return other

    @property
    def alphabet(self):
        """Return the alphabet property."""
        return self._alphabet

    def copy(self):
        """Create and return a copy of the array."""
        other = Array(alphabet=self._alphabet, data=self)
        return other

    def get(self, key, value=None):
        """Return the value of the key if found; return value otherwise."""
        try:
            return self[key]
        except IndexError:
            return value

    def items(self):
        """Return an iterator  of (key, value) pairs in the array."""
        dims = len(self.shape)
        if dims == 1:
            for index, key in enumerate(self._alphabet):
                value = numpy.ndarray.__getitem__(self, index)
                yield key, value
        elif dims == 2:
            for i1, c1 in enumerate(self._alphabet):
                for i2, c2 in enumerate(self._alphabet):
                    key = (c1, c2)
                    value = numpy.ndarray.__getitem__(self, (i1, i2))
                    yield key, value
        else:
            raise RuntimeError("array has unexpected shape %s" % self.shape)

    def keys(self):
        """Return a tuple with the keys associated with the array."""
        dims = len(self.shape)
        alphabet = self._alphabet
        if dims == 1:
            return tuple(alphabet)
        elif dims == 2:
            return tuple((c1, c2) for c2 in alphabet for c1 in alphabet)
        else:
            raise RuntimeError("array has unexpected shape %s" % self.shape)

    def values(self):
        """Return a tuple with the values stored in the array."""
        dims = len(self.shape)
        alphabet = self._alphabet
        if dims == 1:
            return tuple(self)
        elif dims == 2:
            n1, n2 = self.shape
            return tuple(
                numpy.ndarray.__getitem__(self, (i1, i2))
                for i2 in range(n2)
                for i1 in range(n1)
            )
        else:
            raise RuntimeError("array has unexpected shape %s" % self.shape)

    def update(self, E=None, **F):
        """Update the array from dict/iterable E and F."""
        if E is not None:
            try:
                alphabet = E.keys()
            except AttributeError:
                for key, value in E:
                    self[key] = value
            else:
                for key in E:
                    self[key] = E[key]
        for key in F:
            self[key] = F[key]

    def select(self, alphabet):
        """Subset the array by selecting the letters from the specified alphabet."""
        ii = []
        jj = []
        for i, key in enumerate(alphabet):
            try:
                j = self._alphabet.index(key)
            except ValueError:
                continue
            ii.append(i)
            jj.append(j)
        dims = len(self.shape)
        a = Array(alphabet, dims=dims)
        ii = numpy.ix_(*[ii] * dims)
        jj = numpy.ix_(*[jj] * dims)
        a[ii] = numpy.ndarray.__getitem__(self, jj)
        return a

    def _format_1D(self, fmt):
        _alphabet = self._alphabet
        n = len(_alphabet)
        words = [None for i in range(n)]
        lines = []
        try:
            header = self.header
        except AttributeError:
            pass
        else:
            for line in header:
                line = "#  %s\n" % line
                lines.append(line)
        maxwidth = 0
        for i, key in enumerate(_alphabet):
            value = self[key]
            word = fmt % value
            width = len(word)
            if width > maxwidth:
                maxwidth = width
            words[i] = word
        fmt2 = " %" + str(maxwidth) + "s"
        for letter, word in zip(_alphabet, words):
            word = fmt2 % word
            line = letter + word + "\n"
            lines.append(line)
        text = "".join(lines)
        return text

    def _format_2D(self, fmt):
        alphabet = self.alphabet
        n = len(alphabet)
        words = [[None for j in range(n)] for i in range(n)]
        lines = []
        try:
            header = self.header
        except AttributeError:
            pass
        else:
            for line in header:
                line = "#  %s\n" % line
                lines.append(line)
        width = max(len(c) for c in alphabet)
        line = " " * width
        for j, c2 in enumerate(alphabet):
            maxwidth = 0
            for i, c1 in enumerate(alphabet):
                key = (c1, c2)
                value = self[key]
                word = fmt % value
                width = len(word)
                if width > maxwidth:
                    maxwidth = width
                words[i][j] = word
            fmt2 = " %" + str(maxwidth) + "s"
            word = fmt2 % c2
            line += word
            for i, c1 in enumerate(alphabet):
                word = words[i][j]
                words[i][j] = fmt2 % word
        line = line.rstrip() + "\n"
        lines.append(line)
        for letter, row in zip(alphabet, words):
            line = letter + "".join(row) + "\n"
            lines.append(line)
        text = "".join(lines)
        return text

    def __format__(self, fmt):
        if fmt == "":
            if numpy.issubdtype(self.dtype, numpy.integer):
                fmt = "%i"
            else:
                fmt = "%.1f"
        n = len(self.shape)
        if n == 1:
            return self._format_1D(fmt)
        elif n == 2:
            return self._format_2D(fmt)
        else:
            raise RuntimeError("Array has unexpected rank %d" % n)

    def __str__(self):
        return self.__format__("")

    def __repr__(self):
        text = numpy.ndarray.__repr__(self)
        alphabet = self._alphabet
        if isinstance(alphabet, str):
            assert text.endswith(")")
            text = text[:-1] + ",\n         alphabet='%s')" % self._alphabet
        return text


def read(handle, dtype=float):
    """Parse the file and return an Array object."""
    try:
        fp = open(handle)
        lines = fp.readlines()
    except TypeError:
        fp = handle
        try:
            lines = fp.readlines()
        except Exception as e:
            raise e from None
        finally:
            fp.close()
    header = []
    for i, line in enumerate(lines):
        if not line.startswith("#"):
            break
        header.append(line[1:].strip())
    rows = [line.split() for line in lines[i:]]
    if len(rows[0]) == len(rows[1]) == 2:
        alphabet = [key for key, value in rows]
        for key in alphabet:
            if len(key) > 1:
                alphabet = tuple(alphabet)
                break
        else:
            alphabet = "".join(alphabet)
        matrix = Array(alphabet=alphabet, dims=1, dtype=dtype)
        matrix.update(rows)
    else:
        alphabet = rows.pop(0)
        for key in alphabet:
            if len(key) > 1:
                alphabet = tuple(alphabet)
                break
        else:
            alphabet = "".join(alphabet)
        matrix = Array(alphabet=alphabet, dims=2, dtype=dtype)
        for letter1, row in zip(alphabet, rows):
            assert letter1 == row.pop(0)
            for letter2, word in zip(alphabet, row):
                matrix[letter1, letter2] = float(word)
    matrix.header = header
    return matrix


def load(name=None):
    """Load and return a precalculated substitution matrix.

    >>> from Bio.Align import substitution_matrices
    >>> names = substitution_matrices.load()
    """
    path = os.path.realpath(__file__)
    directory = os.path.dirname(path)
    subdirectory = os.path.join(directory, "data")
    if name is None:
        filenames = os.listdir(subdirectory)
        return sorted(filenames)
    path = os.path.join(subdirectory, name)
    matrix = read(path)
    return matrix



# Copyright 2019 by Robert T. Miller.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""PICIO: read and write Protein Internal Coordinate (.pic) data files."""

import re

try:
    import numpy
except ImportError:
    from Bio import MissingPythonDependencyError

    raise MissingPythonDependencyError(
        "Install NumPy to build proteins from internal coordinates."
    )

from Bio.File import as_handle
from Bio.PDB.StructureBuilder import StructureBuilder
from Bio.PDB.parse_pdb_header import _parse_pdb_header_list
from Bio.PDB.PDBExceptions import PDBException

from Bio.PDB.internal_coords import IC_Residue, IC_Chain, Edron, AtomKey

from typing import Dict, TextIO
from Bio.PDB.Structure import Structure


def read_PIC(file: TextIO, verbose: bool = False) -> Structure:
    """Load Protein Internal Coordinate (.pic) data from file.

    PIC file format:
        - comment lines start with #
        - (optional) PDB HEADER record
           - idcode and deposition date recommended but optional
           - deposition date in PDB format or as changed by Biopython
        - (optional) PDB TITLE record
        - repeat:
           - Biopython Residue Full ID - sets residue IDs of returned structure
           - (optional) PDB N, CA, C ATOM records for chain start
           - (optional) PIC Hedra records for residue
           - (optional) PIC Dihedra records for residue
           - (optional) BFAC records listing AtomKeys and b-factors

    An improvement would define relative positions for HOH (water) entries.

    N.B. dihedron (i-1)C-N-CA-CB is ignored in assembly if O exists.

    C-beta is by default placed using O-C-CA-CB, but O is missing
    in some PDB file residues, which means the sidechain cannot be
    placed.  The alternate CB path (i-1)C-N-CA-CB is provided to
    circumvent this, but if this is needed then it must be adjusted in
    conjunction with PHI ((i-1)C-N-CA-C) as they overlap.  (i-1)C-N-CA-CB is
    included by default in .pic files for consistency and informational
    (e.g. statistics gathering) purposes, as otherwise the dihedron would only
    appear in the few cases it is needed for.

    :param Bio.File file: file name or handle
    :param bool verbose: complain when lines not as expected
    :returns: Biopython Structure object, Residues with .internal_coord attributes
        but no coordinates except for chain start N, CA, C atoms if supplied,
        **OR** None on parse fail (silent unless verbose=True)

    """
    pdb_hdr_re = re.compile(
        r"^HEADER\s{4}(?P<cf>.{1,40})"
        r"(?:\s+(?P<dd>\d\d\d\d-\d\d-\d\d|\d\d-\w\w\w-\d\d))?"
        r"(?:\s+(?P<id>[0-9A-Z]{4}))?\s*$"
    )
    # ^\('(?P<pid>\w*)',\s(?P<mdl>\d+),\s'(?P<chn>\w)',\s\('(?P<het>\s|[\w-]+)',\s(?P<pos>\d+),\s'(?P<icode>\s|\w)'\)\)\s(?P<res>[A-Z]{3})\s(\[(?P<segid>[a-zA-z\s]{4})\])?\s*$
    pdb_ttl_re = re.compile(r"^TITLE\s{5}(?P<ttl>.+)\s*$")
    biop_id_re = re.compile(
        r"^\('(?P<pid>[^\s]*)',\s(?P<mdl>\d+),\s"
        r"'(?P<chn>\s|\w)',\s\('(?P<het>\s|[\w\s-]+)"
        r"',\s(?P<pos>-?\d+),\s'(?P<icode>\s|\w)'\)\)"
        r"\s+(?P<res>[\w]{1,3})"
        r"(\s\[(?P<segid>[a-zA-z\s]+)\])?"
        r"\s*$"
    )
    pdb_atm_re = re.compile(
        r"^ATOM\s\s(?:\s*(?P<ser>\d+))\s(?P<atm>[\w\s]{4})"
        r"(?P<alc>\w|\s)(?P<res>[\w]{3})\s(?P<chn>.)"
        r"(?P<pos>[\s\-\d]{4})(?P<icode>[A-Za-z\s])\s\s\s"
        r"(?P<x>[\s\-\d\.]{8})(?P<y>[\s\-\d\.]{8})"
        r"(?P<z>[\s\-\d\.]{8})(?P<occ>[\s\d\.]{6})"
        r"(?P<tfac>[\s\d\.]{6})\s{6}"
        r"(?P<segid>[a-zA-z\s]{4})(?P<elm>.{2})"
        r"(?P<chg>.{2})?\s*$"
    )
    bfac_re = re.compile(
        r"^BFAC:\s([^\s]+\s+[\-\d\.]+)"
        r"\s*([^\s]+\s+[\-\d\.]+)?"
        r"\s*([^\s]+\s+[\-\d\.]+)?"
        r"\s*([^\s]+\s+[\-\d\.]+)?"
        r"\s*([^\s]+\s+[\-\d\.]+)?"
    )
    bfac2_re = re.compile(r"([^\s]+)\s+([\-\d\.]+)")
    struct_builder = StructureBuilder()

    # init empty header dict
    # - could use to parse HEADER and TITLE lines except
    #   deposition_date format changed from original PDB header
    header_dict = _parse_pdb_header_list([])

    curr_SMCS = [None, None, None, None]  # struct model chain seg
    SMCS_init = [
        struct_builder.init_structure,
        struct_builder.init_model,
        struct_builder.init_chain,
        struct_builder.init_seg,
    ]

    sb_res = None

    with as_handle(file, mode="r") as handle:
        for aline in handle.readlines():
            if aline.startswith("#"):
                pass  # skip comment lines
            elif aline.startswith("HEADER "):
                m = pdb_hdr_re.match(aline)
                if m:
                    header_dict["head"] = m.group("cf")  # classification
                    header_dict["idcode"] = m.group("id")
                    header_dict["deposition_date"] = m.group("dd")
                elif verbose:
                    print("Reading pic file", file, "HEADER parse fail: ", aline)
            elif aline.startswith("TITLE "):
                m = pdb_ttl_re.match(aline)
                if m:
                    header_dict["name"] = m.group("ttl").strip()
                    # print('TTL: ', m.group('ttl').strip())
                elif verbose:
                    print("Reading pic file", file, "TITLE parse fail:, ", aline)
            elif aline.startswith("("):  # Biopython ID line for Residue
                m = biop_id_re.match(aline)
                if m:
                    # check SMCS = Structure, Model, Chain, SegID
                    segid = m.group(9)
                    if segid is None:
                        segid = "    "
                    this_SMCS = [m.group(1), int(m.group(2)), m.group(3), segid]
                    if curr_SMCS != this_SMCS:
                        # init new SMCS level as needed
                        for i in range(4):
                            if curr_SMCS[i] != this_SMCS[i]:
                                SMCS_init[i](this_SMCS[i])
                                curr_SMCS[i] = this_SMCS[i]
                                if 0 == i:
                                    # 0 = init structure so add header
                                    struct_builder.set_header(header_dict)
                                elif 1 == i:
                                    # new model means new chain and new segid
                                    curr_SMCS[2] = curr_SMCS[3] = None

                    struct_builder.init_residue(
                        m.group("res"),
                        m.group("het"),
                        int(m.group("pos")),
                        m.group("icode"),
                    )

                    sb_res = struct_builder.residue
                    if 2 == sb_res.is_disordered():
                        for r in sb_res.child_dict.values():
                            if not r.internal_coord:
                                sb_res = r
                                break
                    sb_res.internal_coord = IC_Residue(sb_res)
                    # print('res id:', m.groupdict())
                    # print(report_IC(struct_builder.get_structure()))
                else:
                    if verbose:
                        print(
                            "Reading pic file", file, "residue ID parse fail: ", aline
                        )
                    return None
            elif aline.startswith("ATOM "):
                m = pdb_atm_re.match(aline)
                if m:
                    if sb_res is None:
                        # ATOM without res spec already loaded, not a pic file
                        if verbose:
                            print(
                                "Reading pic file",
                                file,
                                "ATOM without residue configured:, ",
                                aline,
                            )
                        return None
                    if sb_res.resname != m.group("res") or sb_res.id[1] != int(
                        m.group("pos")
                    ):
                        if verbose:
                            print(
                                "Reading pic file",
                                file,
                                "ATOM not in configured residue (",
                                sb_res.resname,
                                str(sb_res.id),
                                "):",
                                aline,
                            )
                        return None
                    coord = numpy.array(
                        (float(m.group("x")), float(m.group("y")), float(m.group("z"))),
                        "f",
                    )
                    struct_builder.init_atom(
                        m.group("atm").strip(),
                        coord,
                        float(m.group("tfac")),
                        float(m.group("occ")),
                        m.group("alc"),
                        m.group("atm"),
                        int(m.group("ser")),
                        m.group("elm").strip(),
                    )

                    # print('atom: ', m.groupdict())
                # elif verbose:
                #     print("Reading pic file", file, "ATOM parse fail:", aline)
            elif aline.startswith("BFAC: "):
                m = bfac_re.match(aline)
                if m:
                    for bfac_pair in m.groups():
                        if bfac_pair is not None:
                            m2 = bfac2_re.match(bfac_pair)
                            if m2 and sb_res is not None and sb_res.internal_coord:
                                rp = sb_res.internal_coord
                                rp.bfactors[m2.group(1)] = float(m2.group(2))
                # else:
                #    print('Reading pic file', file, 'B-factor line fail: ', aline)
            else:
                m = Edron.edron_re.match(aline)
                if m and sb_res is not None:
                    sb_res.internal_coord.load_PIC(m.groupdict())
                elif m:
                    print(
                        "PIC file: ",
                        file,
                        " error: no residue info before reading (di/h)edron data: ",
                        aline,
                    )
                    return None
                elif aline.strip():
                    if verbose:
                        print("Reading PIC file", file, "parse fail on: .", aline, ".")
                    return None

    struct = struct_builder.get_structure()
    for chn in struct.get_chains():
        chnp = chn.internal_coord = IC_Chain(chn)
        # done in IC_Chain init : chnp.set_residues()
        chnp.link_residues()
        chnp.init_edra()

    # print(report_PIC(struct_builder.get_structure()))
    return struct


def _wpr(entity, fp, pdbid, chainid):
    if entity.internal_coord:
        if not chainid or not pdbid:
            chain = entity.parent
            if not chainid:
                chainid = chain.id
            if not pdbid:
                struct = chain.parent.parent
                pdbid = struct.header.get("idcode")

        fp.write(entity.internal_coord.write_PIC(pdbid, chainid))
    else:
        fp.write(IC_Residue._residue_string(entity))


def _enumerate_entity_atoms(entity):
    need = False
    for atm in entity.get_atoms():
        need = not atm.get_serial_number()
        break
    if need:
        anum = 1
        for res in entity.get_residues():
            if 2 == res.is_disordered():
                for r in res.child_dict.values():
                    for atm in r.get_unpacked_list():
                        atm.set_serial_number(anum)
                        anum += 1
            else:
                for atm in res.get_unpacked_list():
                    atm.set_serial_number(anum)
                    anum += 1


def enumerate_atoms(entity):
    """Ensure all atoms in entity have serial_number set."""
    while entity.get_parent():
        entity = entity.get_parent()  # get to top level

    if "S" == entity.level:
        for mdl in entity:  # each model starts with 1
            _enumerate_entity_atoms(mdl)
    else:  # only Chain or Residue, start with 1
        _enumerate_entity_atoms(entity)


def pdb_date(datestr: str) -> str:
    """Convert yyyy-mm-dd date to dd-month-yy."""
    if datestr:
        m = re.match(r"(\d{4})-(\d{2})-(\d{2})", datestr)
        if m:
            mo = [
                "XXX",
                "JAN",
                "FEB",
                "MAR",
                "APR",
                "MAY",
                "JUN",
                "JUL",
                "AUG",
                "SEP",
                "OCT",
                "NOV",
                "DEC",
            ][int(m.group(2))]
            datestr = m.group(3) + "-" + mo + "-" + m.group(1)[-2:]
    return datestr


def write_PIC(entity, file, pdbid=None, chainid=None):
    """Write Protein Internal Coordinates (PIC) to file.

    See read_PIC() for file format.  Recurses to lower entity levels (M, C, R).

    :param Entity entity: Biopython PDB Entity object: S, M, C or R
    :param Bio.File file: file name or handle
    :param str pdbid: PDB idcode, read from entity if not supplied
    :param char chainid: PDB Chain ID, set from C level entity.id if needed
    :raises PDBException: if entity level not S, M, C, or R
    :raises Exception: if entity does not have .level attribute
    """
    enumerate_atoms(entity)

    with as_handle(file, "w") as fp:
        try:
            if "A" == entity.level:
                raise PDBException("No PIC output at Atom level")
            elif "R" == entity.level:
                if 2 == entity.is_disordered():
                    for r in entity.child_dict.values():
                        _wpr(r, fp, pdbid, chainid)
                else:
                    _wpr(entity, fp, pdbid, chainid)
            elif "C" == entity.level:
                if not chainid:
                    chainid = entity.id
                for res in entity:
                    write_PIC(res, fp, pdbid, chainid)
                    pass
            elif "M" == entity.level:
                for chn in entity:
                    write_PIC(chn, fp, pdbid, chainid)
            elif "S" == entity.level:
                if not pdbid:
                    pdbid = entity.header.get("idcode", None)
                hdr = entity.header.get("head", None)
                dd = pdb_date(entity.header.get("deposition_date", None))
                if hdr:
                    fp.write(
                        ("HEADER    {:40}{:8}   {:4}\n").format(
                            hdr.upper(), (dd or ""), (pdbid or "")
                        )
                    )
                nam = entity.header.get("name", None)
                if nam:
                    fp.write("TITLE     " + nam.upper() + "\n")
                for mdl in entity:
                    write_PIC(mdl, fp, pdbid, chainid)
            else:
                raise PDBException("Cannot identify level: " + str(entity.level))
        except KeyError:
            raise Exception(
                "write_PIC: argument is not a Biopython PDB Entity " + str(entity)
            )



# Copyright 2012 by Wibowo Arindrarto.  All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Bio.SearchIO parser for HMMER domain table output format."""

from itertools import chain

from Bio.SearchIO._model import QueryResult, Hit, HSP, HSPFragment

from .hmmer3_tab import Hmmer3TabParser, Hmmer3TabIndexer

__all__ = (
    "Hmmer3DomtabHmmhitParser",
    "Hmmer3DomtabHmmqueryParser",
    "Hmmer3DomtabHmmhitIndexer",
    "Hmmer3DomtabHmmqueryIndexer",
    "Hmmer3DomtabHmmhitWriter",
    "Hmmer3DomtabHmmqueryWriter",
)


class Hmmer3DomtabParser(Hmmer3TabParser):
    """Base hmmer3-domtab iterator."""

    def _parse_row(self):
        """Return a dictionary of parsed row values (PRIVATE)."""
        assert self.line
        cols = [x for x in self.line.strip().split(" ") if x]
        # if len(cols) > 23, we have extra description columns
        # combine them all into one string in the 19th column
        if len(cols) > 23:
            cols[22] = " ".join(cols[22:])
        elif len(cols) < 23:
            cols.append("")
            assert len(cols) == 23

        # assign parsed column data into qresult, hit, and hsp dicts
        qresult = {}
        qresult["id"] = cols[3]  # query name
        qresult["accession"] = cols[4]  # query accession
        qresult["seq_len"] = int(cols[5])  # qlen
        hit = {}
        hit["id"] = cols[0]  # target name
        hit["accession"] = cols[1]  # target accession
        hit["seq_len"] = int(cols[2])  # tlen
        hit["evalue"] = float(cols[6])  # evalue
        hit["bitscore"] = float(cols[7])  # score
        hit["bias"] = float(cols[8])  # bias
        hit["description"] = cols[22]  # description of target
        hsp = {}
        hsp["domain_index"] = int(cols[9])  # # (domain number)
        # not parsing cols[10] since it's basically len(hit)
        hsp["evalue_cond"] = float(cols[11])  # c-evalue
        hsp["evalue"] = float(cols[12])  # i-evalue
        hsp["bitscore"] = float(cols[13])  # score
        hsp["bias"] = float(cols[14])  # bias
        hsp["env_start"] = int(cols[19]) - 1  # env from
        hsp["env_end"] = int(cols[20])  # env to
        hsp["acc_avg"] = float(cols[21])  # acc
        frag = {}
        # strand is always 0, since HMMER now only handles protein
        frag["hit_strand"] = frag["query_strand"] = 0
        frag["hit_start"] = int(cols[15]) - 1  # hmm from
        frag["hit_end"] = int(cols[16])  # hmm to
        frag["query_start"] = int(cols[17]) - 1  # ali from
        frag["query_end"] = int(cols[18])  # ali to
        # HMMER results are always protein
        frag["molecule_type"] = "protein"

        # switch hmm<-->ali coordinates if hmm is not hit
        if not self.hmm_as_hit:
            frag["hit_end"], frag["query_end"] = (frag["query_end"], frag["hit_end"])
            frag["hit_start"], frag["query_start"] = (
                frag["query_start"],
                frag["hit_start"],
            )

        return {"qresult": qresult, "hit": hit, "hsp": hsp, "frag": frag}

    def _parse_qresult(self):
        """Return QueryResult objects (PRIVATE)."""
        # state values, determines what to do for each line
        state_EOF = 0
        state_QRES_NEW = 1
        state_QRES_SAME = 3
        state_HIT_NEW = 2
        state_HIT_SAME = 4
        # dummies for initial states
        qres_state = None
        hit_state = None
        file_state = None
        # dummies for initial id caches
        prev_qid = None
        prev_hid = None
        # dummies for initial parsed value containers
        cur, prev = None, None
        hit_list, hsp_list = [], []
        cur_qid = None
        cur_hid = None
        while True:
            # store previous line's parsed values, for every line after the 1st
            if cur is not None:
                prev = cur
                prev_qid = cur_qid
                prev_hid = cur_hid
            # only parse the line if it's not EOF
            if self.line and not self.line.startswith("#"):
                cur = self._parse_row()
                cur_qid = cur["qresult"]["id"]
                cur_hid = cur["hit"]["id"]
            else:
                file_state = state_EOF
                # mock ID values since the line is empty
                cur_qid, cur_hid = None, None

            # get the state of hit and qresult
            if prev_qid != cur_qid:
                qres_state = state_QRES_NEW
            else:
                qres_state = state_QRES_SAME
            # new hits are hits with different ids or hits in a new qresult
            if prev_hid != cur_hid or qres_state == state_QRES_NEW:
                hit_state = state_HIT_NEW
            else:
                hit_state = state_HIT_SAME

            # start creating objects after the first line (i.e. prev is filled)
            if prev is not None:
                # each line is basically an HSP with one HSPFragment
                frag = HSPFragment(prev_hid, prev_qid)
                for attr, value in prev["frag"].items():
                    setattr(frag, attr, value)
                hsp = HSP([frag])
                for attr, value in prev["hsp"].items():
                    setattr(hsp, attr, value)
                hsp_list.append(hsp)

                # create hit object when we've finished parsing all its hsps
                # i.e. when hit state is state_HIT_NEW
                if hit_state == state_HIT_NEW:
                    hit = Hit(hsp_list)
                    for attr, value in prev["hit"].items():
                        setattr(hit, attr, value)
                    hit_list.append(hit)
                    hsp_list = []

                # create qresult and yield if we're at a new qresult or EOF
                if qres_state == state_QRES_NEW or file_state == state_EOF:
                    qresult = QueryResult(hit_list, prev_qid)
                    for attr, value in prev["qresult"].items():
                        setattr(qresult, attr, value)
                    yield qresult
                    # if current line is EOF, break
                    if file_state == state_EOF:
                        break
                    hit_list = []

            self.line = self.handle.readline()


class Hmmer3DomtabHmmhitParser(Hmmer3DomtabParser):
    """HMMER domain table parser using hit coordinates.

    Parser for the HMMER domain table format that assumes HMM profile
    coordinates are hit coordinates.
    """

    hmm_as_hit = True


class Hmmer3DomtabHmmqueryParser(Hmmer3DomtabParser):
    """HMMER domain table parser using query coordinates.

    Parser for the HMMER domain table format that assumes HMM profile
    coordinates are query coordinates.
    """

    hmm_as_hit = False


class Hmmer3DomtabHmmhitIndexer(Hmmer3TabIndexer):
    """HMMER domain table indexer using hit coordinates.

    Indexer class for HMMER domain table output that assumes HMM profile
    coordinates are hit coordinates.
    """

    _parser = Hmmer3DomtabHmmhitParser
    _query_id_idx = 3


class Hmmer3DomtabHmmqueryIndexer(Hmmer3TabIndexer):
    """HMMER domain table indexer using query coordinates.

    Indexer class for HMMER domain table output that assumes HMM profile
    coordinates are query coordinates.
    """

    _parser = Hmmer3DomtabHmmqueryParser
    _query_id_idx = 3


class Hmmer3DomtabHmmhitWriter:
    """HMMER domain table writer using hit coordinates.

    Writer for hmmer3-domtab output format which writes hit coordinates
    as HMM profile coordinates.
    """

    hmm_as_hit = True

    def __init__(self, handle):
        """Initialize the class."""
        self.handle = handle

    def write_file(self, qresults):
        """Write to the handle.

        Returns a tuple of how many QueryResult, Hit, and HSP objects were written.

        """
        handle = self.handle
        qresult_counter, hit_counter, hsp_counter, frag_counter = 0, 0, 0, 0

        try:
            first_qresult = next(qresults)
        except StopIteration:
            handle.write(self._build_header())
        else:
            # write header
            handle.write(self._build_header(first_qresult))
            # and then the qresults
            for qresult in chain([first_qresult], qresults):
                if qresult:
                    handle.write(self._build_row(qresult))
                    qresult_counter += 1
                    hit_counter += len(qresult)
                    hsp_counter += sum(len(hit) for hit in qresult)
                    frag_counter += sum(len(hit.fragments) for hit in qresult)

        return qresult_counter, hit_counter, hsp_counter, frag_counter

    def _build_header(self, first_qresult=None):
        """Return the header string of a domain HMMER table output (PRIVATE)."""
        # calculate whitespace required
        # adapted from HMMER's source: src/p7_tophits.c#L1157
        if first_qresult:
            # qnamew = max(20, len(first_qresult.id))
            qnamew = 20
            tnamew = max(20, len(first_qresult[0].id))
            try:
                qaccw = max(10, len(first_qresult.acc))
                taccw = max(10, len(first_qresult[0].acc))
            except AttributeError:
                qaccw, taccw = 10, 10
        else:
            qnamew, tnamew, qaccw, taccw = 20, 20, 10, 10
        # Turn black code style off
        # fmt: off
        header = ("#%*s %22s %40s %11s %11s %11s\n"
                  % (tnamew + qnamew - 1 + 15 + taccw + qaccw, "", "--- full sequence ---",
                     "-------------- this domain -------------", "hmm coord",
                     "ali coord", "env coord"))
        header += ("#%-*s %-*s %5s %-*s %-*s %5s %9s %6s %5s %3s %3s %9s "
                   "%9s %6s %5s %5s %5s %5s %5s %5s %5s %4s %s\n"
                   % (tnamew - 1,
                      " target name", taccw, "accession", "tlen", qnamew,
                      "query name", qaccw, "accession", "qlen", "E-value", "score",
                      "bias", "#", "of", "c-Evalue", "i-Evalue", "score", "bias",
                      "from", "to", "from", "to", "from", "to", "acc",
                      "description of target"))
        header += ("#%*s %*s %5s %*s %*s %5s %9s %6s %5s %3s %3s %9s %9s "
                   "%6s %5s %5s %5s %5s %5s %5s %5s %4s %s\n"
                   % (tnamew - 1,
                      "-------------------", taccw, "----------", "-----",
                      qnamew, "--------------------", qaccw, "----------",
                      "-----", "---------", "------", "-----", "---", "---",
                      "---------", "---------", "------", "-----", "-----", "-----",
                      "-----", "-----", "-----", "-----", "----",
                      "---------------------"))
        # Turn black code style on
        # fmt: on
        return header

    def _build_row(self, qresult):
        """Return a string or one row or more of the QueryResult object (PRIVATE)."""
        rows = ""

        # calculate whitespace required
        # adapted from HMMER's source: src/p7_tophits.c#L1083
        qnamew = max(20, len(qresult.id))
        tnamew = max(20, len(qresult[0].id))
        try:
            qaccw = max(10, len(qresult.accession))
            taccw = max(10, len(qresult[0].accession))
            qresult_acc = qresult.accession
        except AttributeError:
            qaccw, taccw = 10, 10
            qresult_acc = "-"

        for hit in qresult:

            # try to get hit accession
            try:
                hit_acc = hit.accession
            except AttributeError:
                hit_acc = "-"

            for hsp in hit.hsps:
                if self.hmm_as_hit:
                    hmm_to = hsp.hit_end
                    hmm_from = hsp.hit_start + 1
                    ali_to = hsp.query_end
                    ali_from = hsp.query_start + 1
                else:
                    hmm_to = hsp.query_end
                    hmm_from = hsp.query_start + 1
                    ali_to = hsp.hit_end
                    ali_from = hsp.hit_start + 1

                rows += (
                    "%-*s %-*s %5d %-*s %-*s %5d %9.2g %6.1f %5.1f %3d"
                    " %3d %9.2g %9.2g %6.1f %5.1f %5d %5d %5ld %5ld"
                    " %5d %5d %4.2f %s\n"
                    % (
                        tnamew,
                        hit.id,
                        taccw,
                        hit_acc,
                        hit.seq_len,
                        qnamew,
                        qresult.id,
                        qaccw,
                        qresult_acc,
                        qresult.seq_len,
                        hit.evalue,
                        hit.bitscore,
                        hit.bias,
                        hsp.domain_index,
                        len(hit.hsps),
                        hsp.evalue_cond,
                        hsp.evalue,
                        hsp.bitscore,
                        hsp.bias,
                        hmm_from,
                        hmm_to,
                        ali_from,
                        ali_to,
                        hsp.env_start + 1,
                        hsp.env_end,
                        hsp.acc_avg,
                        hit.description,
                    )
                )

        return rows


class Hmmer3DomtabHmmqueryWriter(Hmmer3DomtabHmmhitWriter):
    """HMMER domain table writer using query coordinates.

    Writer for hmmer3-domtab output format which writes query coordinates
    as HMM profile coordinates.
    """

    hmm_as_hit = False


# if not used as a module, run the doctest
if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest()



# Copyright 2001-2009 Brad Chapman.
# Revisions copyright 2009-2016 by Peter Cock.
# Revisions copyright 2009 by David Winter.
# Revisions copyright 2009-2010 by Leighton Pritchard.
# All rights reserved.
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.
"""Code to interact with and run various EMBOSS programs.

These classes follow the AbstractCommandline interfaces for running
programs.
"""


from Bio.Application import _Option, _Switch, AbstractCommandline


class _EmbossMinimalCommandLine(AbstractCommandline):
    """Base Commandline object for EMBOSS wrappers (PRIVATE).

    This is provided for subclassing, it deals with shared options
    common to all the EMBOSS tools:

    Attributes:
     - auto               Turn off prompts
     - stdout             Write standard output
     - filter             Read standard input, write standard output
     - options            Prompt for standard and additional values
     - debug              Write debug output to program.dbg
     - verbose            Report some/full command line options
     - help               Report command line options. More
                          information on associated and general
                          qualifiers can be found with -help -verbose
     - warning            Report warnings
     - error              Report errors
     - fatal              Report fatal errors
     - die                Report dying program messages

    """

    def __init__(self, cmd=None, **kwargs):
        assert cmd is not None
        extra_parameters = [
            _Switch(
                ["-auto", "auto"],
                "Turn off prompts.\n\n"
                "Automatic mode disables prompting, so we recommend you set this "
                "argument all the time when calling an EMBOSS tool from Biopython.",
            ),
            _Switch(["-stdout", "stdout"], "Write standard output."),
            _Switch(
                ["-filter", "filter"], "Read standard input, write standard output."
            ),
            _Switch(
                ["-options", "options"],
                "Prompt for standard and additional values.\n\n"
                "If you are calling an EMBOSS tool from within Biopython, "
                "we DO NOT recommend using this option.",
            ),
            _Switch(["-debug", "debug"], "Write debug output to program.dbg."),
            _Switch(["-verbose", "verbose"], "Report some/full command line options"),
            _Switch(
                ["-help", "help"],
                "Report command line options.\n\n"
                "More information on associated and general qualifiers "
                "can be found with -help -verbose",
            ),
            _Switch(["-warning", "warning"], "Report warnings."),
            _Switch(["-error", "error"], "Report errors."),
            _Switch(["-die", "die"], "Report dying program messages."),
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        AbstractCommandline.__init__(self, cmd, **kwargs)


class _EmbossCommandLine(_EmbossMinimalCommandLine):
    """Base Commandline object for EMBOSS wrappers (PRIVATE).

    This is provided for subclassing, it deals with shared options
    common to all the EMBOSS tools plus:

     - outfile            Output filename

    """

    def __init__(self, cmd=None, **kwargs):
        assert cmd is not None
        extra_parameters = [
            _Option(["-outfile", "outfile"], "Output filename", filename=True)
        ]
        try:
            # Insert extra parameters - at the start just in case there
            # are any arguments which must come last:
            self.parameters = extra_parameters + self.parameters
        except AttributeError:
            # Should we raise an error?  The subclass should have set this up!
            self.parameters = extra_parameters
        _EmbossMinimalCommandLine.__init__(self, cmd, **kwargs)

    def _validate(self):
        # Check the outfile, filter, or stdout option has been set.
        # We can't simply do this via the required flag for the outfile
        # output - this seems the simplest solution.
        if not (self.outfile or self.filter or self.stdout):
            raise ValueError(
                "You must either set outfile (output filename), "
                "or enable filter or stdout (output to stdout)."
            )
        return _EmbossMinimalCommandLine._validate(self)


class Primer3Commandline(_EmbossCommandLine):
    """Commandline object for the Primer3 interface from EMBOSS.

    The precise set of supported arguments depends on your version of EMBOSS.
    This version accepts arguments current at EMBOSS 6.1.0:

    >>> cline = Primer3Commandline(sequence="mysequence.fas", auto=True, hybridprobe=True)
    >>> cline.explainflag = True
    >>> cline.osizeopt=20
    >>> cline.psizeopt=200
    >>> cline.outfile = "myresults.out"
    >>> cline.bogusparameter = 1967  # Invalid parameter
    Traceback (most recent call last):
        ...
    ValueError: Option name bogusparameter was not found.
    >>> print(cline)
    eprimer3 -auto -outfile=myresults.out -sequence=mysequence.fas -hybridprobe=True -psizeopt=200 -osizeopt=20 -explainflag=True

    """

    def __init__(self, cmd="eprimer3", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "Sequence to choose primers from.",
                is_required=True,
            ),
            _Option(["-task", "task"], "Tell eprimer3 what task to perform."),
            _Option(
                ["-hybridprobe", "hybridprobe"],
                "Find an internal oligo to use as a hyb probe.",
            ),
            _Option(
                ["-numreturn", "numreturn"], "Maximum number of primer pairs to return."
            ),
            _Option(
                ["-includedregion", "includedregion"],
                "Subregion of the sequence in which to pick primers.",
            ),
            _Option(["-target", "target"], "Sequence to target for flanking primers."),
            _Option(
                ["-excludedregion", "excludedregion"],
                "Regions to exclude from primer picking.",
            ),
            _Option(
                ["-forwardinput", "forwardinput"],
                "Sequence of a forward primer to check.",
            ),
            _Option(
                ["-reverseinput", "reverseinput"],
                "Sequence of a reverse primer to check.",
            ),
            _Option(
                ["-gcclamp", "gcclamp"],
                "The required number of Gs and Cs at the 3' of each primer.",
            ),
            _Option(["-osize", "osize"], "Optimum length of a primer oligo."),
            _Option(["-minsize", "minsize"], "Minimum length of a primer oligo."),
            _Option(["-maxsize", "maxsize"], "Maximum length of a primer oligo."),
            _Option(
                ["-otm", "otm"],
                "Melting temperature for primer oligo (OBSOLETE).\n\n"
                "Option replaced in EMBOSS 6.6.0 by -opttm",
            ),
            _Option(
                ["-opttm", "opttm"],
                "Optimum melting temperature for a primer oligo.\n\n"
                "Option added in EMBOSS 6.6.0, replacing -otm",
            ),
            _Option(
                ["-mintm", "mintm"], "Minimum melting temperature for a primer oligo."
            ),
            _Option(
                ["-maxtm", "maxtm"], "Maximum melting temperature for a primer oligo."
            ),
            _Option(
                ["-maxdifftm", "maxdifftm"],
                "Maximum difference in melting temperatures between "
                "forward and reverse primers.",
            ),
            _Option(["-ogcpercent", "ogcpercent"], "Optimum GC% for a primer."),
            _Option(["-mingc", "mingc"], "Minimum GC% for a primer."),
            _Option(["-maxgc", "maxgc"], "Maximum GC% for a primer."),
            _Option(
                ["-saltconc", "saltconc"], "Millimolar salt concentration in the PCR."
            ),
            _Option(
                ["-dnaconc", "dnaconc"],
                "Nanomolar concentration of annealing oligos in the PCR.",
            ),
            _Option(
                ["-maxpolyx", "maxpolyx"],
                "Maximum allowable mononucleotide repeat length in a primer.",
            ),
            # Primer length:
            _Option(["-psizeopt", "psizeopt"], "Optimum size for the PCR product."),
            _Option(
                ["-prange", "prange"], "Acceptable range of length for the PCR product."
            ),
            # Primer temperature:
            _Option(
                ["-ptmopt", "ptmopt"],
                "Optimum melting temperature for the PCR product.",
            ),
            _Option(
                ["-ptmmin", "ptmmin"],
                "Minimum allowed melting temperature for the amplicon.",
            ),
            _Option(
                ["-ptmmax", "ptmmax"],
                "Maximum allowed melting temperature for the amplicon.",
            ),
            # Note to self, should be -oexcludedregion not -oexcluderegion
            _Option(
                ["-oexcludedregion", "oexcludedregion"],
                "Do not pick internal oligos in this region.",
            ),
            _Option(["-oligoinput", "oligoinput"], "Sequence of the internal oligo."),
            # Oligo length:
            _Option(["-osizeopt", "osizeopt"], "Optimum length of internal oligo."),
            _Option(["-ominsize", "ominsize"], "Minimum length of internal oligo."),
            _Option(["-omaxsize", "omaxsize"], "Maximum length of internal oligo."),
            # Oligo GC temperature:
            _Option(
                ["-otmopt", "otmopt"], "Optimum melting temperature of internal oligo."
            ),
            _Option(
                ["-otmmin", "otmmin"], "Minimum melting temperature of internal oligo."
            ),
            _Option(
                ["-otmmax", "otmmax"], "Maximum melting temperature of internal oligo."
            ),
            # Oligo GC percent:
            _Option(["-ogcopt", "ogcopt"], "Optimum GC% for internal oligo."),
            _Option(["-ogcmin", "ogcmin"], "Minimum GC% for internal oligo."),
            _Option(["-ogcmax", "ogcmax"], "Maximum GC% for internal oligo."),
            # Oligo salt concentration:
            _Option(
                ["-osaltconc", "osaltconc"],
                "Millimolar concentration of salt in the hybridisation.",
            ),
            _Option(
                ["-odnaconc", "odnaconc"],
                "Nanomolar concentration of internal oligo in the hybridisation.",
            ),
            # Oligo self complementarity
            _Option(
                ["-oanyself", "oanyself"],
                "Maximum allowable alignment score for self-complementarity.",
            ),
            _Option(
                ["-oendself", "oendself"],
                "Max 3'-anchored self-complementarity global alignment score.",
            ),
            _Option(
                ["-opolyxmax", "opolyxmax"],
                "Maximum length of mononucleotide repeat in internal oligo.",
            ),
            _Option(
                ["-mispriminglibraryfile", "mispriminglibraryfile"],
                "File containing library of sequences to avoid amplifying",
            ),
            _Option(
                ["-maxmispriming", "maxmispriming"],
                "Maximum allowed similarity of primers to sequences in "
                "library specified by -mispriminglibrary",
            ),
            _Option(
                ["-omishybmax", "omishybmax"],
                "Maximum alignment score for hybridisation of internal oligo to "
                "library specified by -mishyblibraryfile.",
            ),
            _Option(
                ["-mishyblibraryfile", "mishyblibraryfile"],
                "Library file of seqs to avoid internal oligo hybridisation.",
            ),
            _Option(
                ["-explainflag", "explainflag"],
                "Produce output tags with eprimer3 statistics",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class PrimerSearchCommandline(_EmbossCommandLine):
    """Commandline object for the primersearch program from EMBOSS."""

    def __init__(self, cmd="primersearch", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-seqall", "-sequences", "sequences", "seqall"],
                "Sequence to look for the primer pairs in.",
                is_required=True,
            ),
            # When this wrapper was written primersearch used -sequences
            # as the argument name. Since at least EMBOSS 5.0 (and
            # perhaps earlier) this has been -seqall instead.
            _Option(
                ["-infile", "-primers", "primers", "infile"],
                "File containing the primer pairs to search for.",
                filename=True,
                is_required=True,
            ),
            # When this wrapper was written primersearch used -primers
            # as the argument name. Since at least EMBOSS 5.0 (and
            # perhaps earlier) this has been -infile instead.
            _Option(
                ["-mismatchpercent", "mismatchpercent"],
                "Allowed percentage mismatch (any integer value, default 0).",
                is_required=True,
            ),
            _Option(
                ["-snucleotide", "snucleotide"], "Sequences are nucleotide (boolean)"
            ),
            _Option(["-sprotein", "sprotein"], "Sequences are protein (boolean)"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FDNADistCommandline(_EmbossCommandLine):
    """Commandline object for the fdnadist program from EMBOSS.

    fdnadist is an EMBOSS wrapper for the PHYLIP program dnadist for
    calulating distance matrices from DNA sequence files.
    """

    def __init__(self, cmd="fdnadist", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "seq file to use (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(["-method", "method"], "sub. model [f,k,j,l,s]", is_required=True),
            _Option(["-gamma", "gamma"], "gamma [g, i,n]"),
            _Option(
                ["-ncategories", "ncategories"], "number of rate catergories (1-9)"
            ),
            _Option(["-rate", "rate"], "rate for each category"),
            _Option(
                ["-categories", "categories"], "File of substitution rate categories"
            ),
            _Option(["-weights", "weights"], "weights file"),
            _Option(
                ["-gammacoefficient", "gammacoefficient"], "value for gamma (> 0.001)"
            ),
            _Option(["-invarfrac", "invarfrac"], "proportoin of invariant sites"),
            _Option(["-ttratio", "ttratio"], "ts/tv ratio"),
            _Option(["-freqsfrom", "freqsfrom"], "use emprical base freqs"),
            _Option(["-basefreq", "basefreq"], "specify basefreqs"),
            _Option(["-lower", "lower"], "lower triangle matrix (y/N)"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FTreeDistCommandline(_EmbossCommandLine):
    """Commandline object for the ftreedist program from EMBOSS.

    ftreedist is an EMBOSS wrapper for the PHYLIP program treedist used for
    calulating distance measures between phylogentic trees.
    """

    def __init__(self, cmd="ftreedist", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-intreefile", "intreefile"],
                "tree file to score (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(["-dtype", "dtype"], "distance type ([S]ymetric, [b]ranch score)"),
            _Option(
                ["-pairing", "pairing"],
                "tree pairing method ([A]djacent pairs, all [p]ossible pairs)",
            ),
            _Option(["-style", "style"], "output style - [V]erbose, [f]ill, [s]parse"),
            _Option(["-noroot", "noroot"], "treat trees as rooted [N/y]"),
            _Option(
                ["-outgrno", "outgrno"],
                "which taxon to root the trees with (starts from 0)",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FNeighborCommandline(_EmbossCommandLine):
    """Commandline object for the fneighbor program from EMBOSS.

    fneighbor is an EMBOSS wrapper for the PHYLIP program neighbor used for
    calulating neighbor-joining or UPGMA trees from distance matrices.
    """

    def __init__(self, cmd="fneighbor", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-datafile", "datafile"],
                "dist file to use (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-matrixtype", "matrixtype"],
                "is martrix [S]quare pr [u]pper or [l]ower",
            ),
            _Option(["-treetype", "treetype"], "nj or UPGMA tree (n/u)"),
            _Option(["-outgrno", "outgrno"], "taxon to use as OG"),
            _Option(["-jumble", "jumble"], "randommise input order (Y/n)"),
            _Option(["-seed", "seed"], "provide a random seed"),
            _Option(["-trout", "trout"], "write tree (Y/n)"),
            _Option(["-outtreefile", "outtreefile"], "filename for output tree"),
            _Option(["-progress", "progress"], "print progress (Y/n)"),
            _Option(["-treeprint", "treeprint"], "print tree (Y/n)"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FSeqBootCommandline(_EmbossCommandLine):
    """Commandline object for the fseqboot program from EMBOSS.

    fseqboot is an EMBOSS wrapper for the PHYLIP program seqboot used to
    pseudo-sample alignment files.
    """

    def __init__(self, cmd="fseqboot", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "seq file to sample (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(["-categories", "catergories"], "file of input categories"),
            _Option(["-weights", "weights"], " weights file"),
            _Option(["-test", "test"], "specify operation, default is bootstrap"),
            _Option(["-regular", "regular"], "absolute number to resample"),
            _Option(["-fracsample", "fracsample"], "fraction to resample"),
            _Option(
                ["-rewriteformat", "rewriteformat"],
                "output format ([P]hyilp, [n]exus, [x]ml",
            ),
            _Option(["-seqtype", "seqtype"], "output format ([D]na, [p]rotein, [r]na"),
            _Option(["-blocksize", "blocksize"], "print progress (Y/n)"),
            _Option(["-reps", "reps"], "how many replicates, defaults to 100)"),
            _Option(
                ["-justweights", "jusweights"],
                "what to write out [D]atasets of just [w]eights",
            ),
            _Option(["-seed", "seed"], "specify random seed"),
            _Option(["-dotdiff", "dotdiff"], "Use dot-differencing? [Y/n]"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FDNAParsCommandline(_EmbossCommandLine):
    """Commandline object for the fdnapars program from EMBOSS.

    fdnapars is an EMBOSS version of the PHYLIP program dnapars, for
    estimating trees from DNA sequences using parsiomny. Calling this command
    without providing a value for the option "-intreefile" will invoke
    "interactive mode" (and as a result fail if called with subprocess) if
    "-auto" is not set to true.
    """

    def __init__(self, cmd="fdnapars", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "seq file to use (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(["-intreefile", "intreefile"], "Phylip tree file"),
            _Option(["-weights", "weights"], "weights file"),
            _Option(["-maxtrees", "maxtrees"], "max trees to save during run"),
            _Option(["-thorough", "thorough"], "more thorough search (Y/n)"),
            _Option(["-rearrange", "rearrange"], "Rearrange on just 1 best tree (Y/n)"),
            _Option(
                ["-transversion", "transversion"], "Use tranversion parsimony (y/N)"
            ),
            _Option(
                ["-njumble", "njumble"],
                "number of times to randomise input order (default is 0)",
            ),
            _Option(["-seed", "seed"], "provide random seed"),
            _Option(["-outgrno", "outgrno"], "Specify outgroup"),
            _Option(["-thresh", "thresh"], "Use threshold parsimony (y/N)"),
            _Option(["-threshold", "threshold"], "Threshold value"),
            _Option(["-trout", "trout"], "Write trees to file (Y/n)"),
            _Option(["-outtreefile", "outtreefile"], "filename for output tree"),
            _Option(["-dotdiff", "dotdiff"], "Use dot-differencing? [Y/n]"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FProtParsCommandline(_EmbossCommandLine):
    """Commandline object for the fdnapars program from EMBOSS.

    fprotpars is an EMBOSS version of the PHYLIP program protpars, for
    estimating trees from protein  sequences using parsiomny. Calling this
    command without providing a value for the option "-intreefile" will invoke
    "interactive mode" (and as a result fail if called with subprocess) if
    "-auto" is not set to true.
    """

    def __init__(self, cmd="fprotpars", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "seq file to use (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(["-intreefile", "intreefile"], "Phylip tree file to score"),
            _Option(
                ["-outtreefile", "outtreefile"],
                "phylip tree output file",
                filename=True,
                is_required=True,
            ),
            _Option(["-weights", "weights"], "weights file"),
            _Option(["-whichcode", "whichcode"], "which genetic code, [U,M,V,F,Y]]"),
            _Option(
                ["-njumble", "njumble"],
                "number of times to randomise input order (default is 0)",
            ),
            _Option(["-seed", "seed"], "provide random seed"),
            _Option(["-outgrno", "outgrno"], "Specify outgroup"),
            _Option(["-thresh", "thresh"], "Use threshold parsimony (y/N)"),
            _Option(["-threshold", "threshold"], "Threshold value"),
            _Option(["-trout", "trout"], "Write trees to file (Y/n)"),
            _Option(["-dotdiff", "dotdiff"], "Use dot-differencing? [Y/n]"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FProtDistCommandline(_EmbossCommandLine):
    """Commandline object for the fprotdist program from EMBOSS.

    fprotdist is an EMBOSS wrapper for the PHYLIP program protdist used to
    estimate trees from protein sequences using parsimony
    """

    def __init__(self, cmd="fprotdist", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "seq file to use (phylip)",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-ncategories", "ncategories"], "number of rate catergories (1-9)"
            ),
            _Option(["-rate", "rate"], "rate for each category"),
            _Option(["-catergories", "catergories"], "file of rates"),
            _Option(["-weights", "weights"], "weights file"),
            _Option(["-method", "method"], "sub. model [j,h,d,k,s,c]"),
            _Option(["-gamma", "gamma"], "gamma [g, i,c]"),
            _Option(
                ["-gammacoefficient", "gammacoefficient"], "value for gamma (> 0.001)"
            ),
            _Option(
                ["-invarcoefficient", "invarcoefficient"],
                "float for variation of substitution rate among sites",
            ),
            _Option(["-aacateg", "aacateg"], "Choose the category to use [G,C,H]"),
            _Option(["-whichcode", "whichcode"], "genetic code [c,m,v,f,y]"),
            _Option(["-ease", "ease"], "Pob change catergory (float between -0 and 1)"),
            _Option(["-ttratio", "ttratio"], "Transition/transversion ratio (0-1)"),
            _Option(
                ["-basefreq", "basefreq"], "DNA base frequencies (space separated list)"
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FConsenseCommandline(_EmbossCommandLine):
    """Commandline object for the fconsense program from EMBOSS.

    fconsense is an EMBOSS wrapper for the PHYLIP program consense used to
    calculate consensus trees.
    """

    def __init__(self, cmd="fconsense", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-intreefile", "intreefile"],
                "file with phylip trees to make consensus from",
                filename=True,
                is_required=True,
            ),
            _Option(["-method", "method"], "consensus method [s, mr, MRE, ml]"),
            _Option(
                ["-mlfrac", "mlfrac"],
                "cut-off freq for branch to appear in consensus (0.5-1.0)",
            ),
            _Option(["-root", "root"], "treat trees as rooted (YES, no)"),
            _Option(["-outgrno", "outgrno"], "OTU to use as outgroup (starts from 0)"),
            _Option(["-trout", "trout"], "treat trees as rooted (YES, no)"),
            _Option(
                ["-outtreefile", "outtreefile"], "Phylip tree output file (optional)"
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class WaterCommandline(_EmbossCommandLine):
    """Commandline object for the water program from EMBOSS."""

    def __init__(self, cmd="water", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-asequence", "asequence"],
                "First sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-bsequence", "bsequence"],
                "Second sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(["-gapopen", "gapopen"], "Gap open penalty", is_required=True),
            _Option(
                ["-gapextend", "gapextend"], "Gap extension penalty", is_required=True
            ),
            _Option(["-datafile", "datafile"], "Matrix file", filename=True),
            _Switch(
                ["-nobrief", "nobrief"], "Display extended identity and similarity"
            ),
            _Switch(["-brief", "brief"], "Display brief identity and similarity"),
            _Option(
                ["-similarity", "similarity"], "Display percent identity and similarity"
            ),
            _Option(
                ["-snucleotide", "snucleotide"], "Sequences are nucleotide (boolean)"
            ),
            _Option(["-sprotein", "sprotein"], "Sequences are protein (boolean)"),
            _Option(
                ["-aformat", "aformat"],
                "Display output in a different specified output format",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class NeedleCommandline(_EmbossCommandLine):
    """Commandline object for the needle program from EMBOSS."""

    def __init__(self, cmd="needle", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-asequence", "asequence"],
                "First sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-bsequence", "bsequence"],
                "Second sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(["-gapopen", "gapopen"], "Gap open penalty", is_required=True),
            _Option(
                ["-gapextend", "gapextend"], "Gap extension penalty", is_required=True
            ),
            _Option(["-datafile", "datafile"], "Matrix file", filename=True),
            _Option(["-endweight", "endweight"], "Apply And gap penalties"),
            _Option(
                ["-endopen", "endopen"],
                "The score taken away when an end gap is created.",
            ),
            _Option(
                ["-endextend", "endextend"],
                "The score added to the end gap penality for each base or "
                "residue in the end gap.",
            ),
            _Switch(
                ["-nobrief", "nobrief"], "Display extended identity and similarity"
            ),
            _Switch(["-brief", "brief"], "Display brief identity and similarity"),
            _Option(
                ["-similarity", "similarity"], "Display percent identity and similarity"
            ),
            _Option(
                ["-snucleotide", "snucleotide"], "Sequences are nucleotide (boolean)"
            ),
            _Option(["-sprotein", "sprotein"], "Sequences are protein (boolean)"),
            _Option(
                ["-aformat", "aformat"],
                "Display output in a different specified output format",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class NeedleallCommandline(_EmbossCommandLine):
    """Commandline object for the needleall program from EMBOSS."""

    def __init__(self, cmd="needleall", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-asequence", "asequence"],
                "First sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-bsequence", "bsequence"],
                "Second sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(["-gapopen", "gapopen"], "Gap open penalty", is_required=True),
            _Option(
                ["-gapextend", "gapextend"], "Gap extension penalty", is_required=True
            ),
            _Option(["-datafile", "datafile"], "Matrix file", filename=True),
            _Option(
                ["-minscore", "minscore"],
                "Exclude alignments with scores below this threshold score.",
            ),
            _Option(["-errorfile", "errorfile"], "Error file to be written to."),
            _Option(["-endweight", "endweight"], "Apply And gap penalties"),
            _Option(
                ["-endopen", "endopen"],
                "The score taken away when an end gap is created.",
            ),
            _Option(
                ["-endextend", "endextend"],
                "The score added to the end gap penality for each base or "
                "residue in the end gap.",
            ),
            _Switch(
                ["-nobrief", "nobrief"], "Display extended identity and similarity"
            ),
            _Switch(["-brief", "brief"], "Display brief identity and similarity"),
            _Option(
                ["-similarity", "similarity"], "Display percent identity and similarity"
            ),
            _Option(
                ["-snucleotide", "snucleotide"], "Sequences are nucleotide (boolean)"
            ),
            _Option(["-sprotein", "sprotein"], "Sequences are protein (boolean)"),
            _Option(
                ["-aformat", "aformat"],
                "Display output in a different specified output format",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class StretcherCommandline(_EmbossCommandLine):
    """Commandline object for the stretcher program from EMBOSS."""

    def __init__(self, cmd="stretcher", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-asequence", "asequence"],
                "First sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-bsequence", "bsequence"],
                "Second sequence to align",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-gapopen", "gapopen"],
                "Gap open penalty",
                is_required=True,
                checker_function=lambda value: isinstance(value, int),
            ),
            _Option(
                ["-gapextend", "gapextend"],
                "Gap extension penalty",
                is_required=True,
                checker_function=lambda value: isinstance(value, int),
            ),
            _Option(["-datafile", "datafile"], "Matrix file", filename=True),
            _Option(
                ["-snucleotide", "snucleotide"], "Sequences are nucleotide (boolean)"
            ),
            _Option(["-sprotein", "sprotein"], "Sequences are protein (boolean)"),
            _Option(
                ["-aformat", "aformat"],
                "Display output in a different specified output format",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FuzznucCommandline(_EmbossCommandLine):
    """Commandline object for the fuzznuc program from EMBOSS."""

    def __init__(self, cmd="fuzznuc", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"], "Sequence database USA", is_required=True
            ),
            _Option(
                ["-pattern", "pattern"],
                "Search pattern, using standard IUPAC one-letter codes",
                is_required=True,
            ),
            _Option(["-pmismatch", "pmismatch"], "Number of mismatches"),
            _Option(["-complement", "complement"], "Search complementary strand"),
            _Option(["-rformat", "rformat"], "Specify the report format to output in."),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class FuzzproCommandline(_EmbossCommandLine):
    """Commandline object for the fuzzpro program from EMBOSS."""

    def __init__(self, cmd="fuzzpro", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"], "Sequence database USA", is_required=True
            ),
            _Option(
                ["-pattern", "pattern"],
                "Search pattern, using standard IUPAC one-letter codes",
                is_required=True,
            ),
            _Option(["-pmismatch", "pmismatch"], "Number of mismatches"),
            _Option(["-rformat", "rformat"], "Specify the report format to output in."),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class Est2GenomeCommandline(_EmbossCommandLine):
    """Commandline object for the est2genome program from EMBOSS."""

    def __init__(self, cmd="est2genome", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(["-est", "est"], "EST sequence(s)", is_required=True),
            _Option(["-genome", "genome"], "Genomic sequence", is_required=True),
            _Option(["-match", "match"], "Score for matching two bases"),
            _Option(["-mismatch", "mismatch"], "Cost for mismatching two bases"),
            _Option(
                ["-gappenalty", "gappenalty"],
                "Cost for deleting a single base in either sequence, "
                "excluding introns",
            ),
            _Option(
                ["-intronpenalty", "intronpenalty"],
                "Cost for an intron, independent of length.",
            ),
            _Option(
                ["-splicepenalty", "splicepenalty"],
                "Cost for an intron, independent of length "
                "and starting/ending on donor-acceptor sites",
            ),
            _Option(
                ["-minscore", "minscore"],
                "Exclude alignments with scores below this threshold score.",
            ),
            _Option(
                ["-reverse", "reverse"], "Reverse the orientation of the EST sequence"
            ),
            _Option(["-splice", "splice"], "Use donor and acceptor splice sites."),
            _Option(
                ["-mode", "mode"],
                "This determines the comparion mode. 'both', 'forward', or 'reverse'",
            ),
            _Option(
                ["-best", "best"],
                "You can print out all comparisons instead of just the best",
            ),
            _Option(["-space", "space"], "for linear-space recursion."),
            _Option(["-shuffle", "shuffle"], "Shuffle"),
            _Option(["-seed", "seed"], "Random number seed"),
            _Option(["-align", "align"], "Show the alignment."),
            _Option(["-width", "width"], "Alignment width"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class ETandemCommandline(_EmbossCommandLine):
    """Commandline object for the etandem program from EMBOSS."""

    def __init__(self, cmd="etandem", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"], "Sequence", filename=True, is_required=True
            ),
            _Option(
                ["-minrepeat", "minrepeat"], "Minimum repeat size", is_required=True
            ),
            _Option(
                ["-maxrepeat", "maxrepeat"], "Maximum repeat size", is_required=True
            ),
            _Option(["-threshold", "threshold"], "Threshold score"),
            _Option(["-mismatch", "mismatch"], "Allow N as a mismatch"),
            _Option(["-uniform", "uniform"], "Allow uniform consensus"),
            _Option(["-rformat", "rformat"], "Output report format"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class EInvertedCommandline(_EmbossCommandLine):
    """Commandline object for the einverted program from EMBOSS."""

    def __init__(self, cmd="einverted", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"], "Sequence", filename=True, is_required=True
            ),
            _Option(["-gap", "gap"], "Gap penalty", filename=True, is_required=True),
            _Option(
                ["-threshold", "threshold"], "Minimum score threshold", is_required=True
            ),
            _Option(["-match", "match"], "Match score", is_required=True),
            _Option(["-mismatch", "mismatch"], "Mismatch score", is_required=True),
            _Option(
                ["-maxrepeat", "maxrepeat"],
                "Maximum separation between the start and end of repeat",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class PalindromeCommandline(_EmbossCommandLine):
    """Commandline object for the palindrome program from EMBOSS."""

    def __init__(self, cmd="palindrome", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"], "Sequence", filename=True, is_required=True
            ),
            _Option(
                ["-minpallen", "minpallen"],
                "Minimum palindrome length",
                is_required=True,
            ),
            _Option(
                ["-maxpallen", "maxpallen"],
                "Maximum palindrome length",
                is_required=True,
            ),
            _Option(
                ["-gaplimit", "gaplimit"],
                "Maximum gap between repeats",
                is_required=True,
            ),
            _Option(
                ["-nummismatches", "nummismatches"],
                "Number of mismatches allowed",
                is_required=True,
            ),
            _Option(
                ["-overlap", "overlap"], "Report overlapping matches", is_required=True
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class TranalignCommandline(_EmbossCommandLine):
    """Commandline object for the tranalign program from EMBOSS."""

    def __init__(self, cmd="tranalign", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-asequence", "asequence"],
                "Nucleotide sequences to be aligned.",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-bsequence", "bsequence"],
                "Protein sequence alignment",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-outseq", "outseq"],
                "Output sequence file.",
                filename=True,
                is_required=True,
            ),
            _Option(["-table", "table"], "Code to use"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class DiffseqCommandline(_EmbossCommandLine):
    """Commandline object for the diffseq program from EMBOSS."""

    def __init__(self, cmd="diffseq", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-asequence", "asequence"],
                "First sequence to compare",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-bsequence", "bsequence"],
                "Second sequence to compare",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-wordsize", "wordsize"],
                "Word size to use for comparisons (10 default)",
                is_required=True,
            ),
            _Option(
                ["-aoutfeat", "aoutfeat"],
                "File for output of first sequence's features",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-boutfeat", "boutfeat"],
                "File for output of second sequence's features",
                filename=True,
                is_required=True,
            ),
            _Option(["-rformat", "rformat"], "Output report file format"),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


class IepCommandline(_EmbossCommandLine):
    """Commandline for EMBOSS iep: calculated isoelectric point and charge.

    Examples
    --------
    >>> from Bio.Emboss.Applications import IepCommandline
    >>> iep_cline = IepCommandline(sequence="proteins.faa",
    ...                            outfile="proteins.txt")
    >>> print(iep_cline)
    iep -outfile=proteins.txt -sequence=proteins.faa

    You would typically run the command line with iep_cline() or via the
    Python subprocess module, as described in the Biopython tutorial.

    """

    def __init__(self, cmd="iep", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "Protein sequence(s) filename",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-amino", "amino"],
                """Number of N-termini

                    Integer 0 (default) or more.
                    """,
            ),
            _Option(
                ["-carboxyl", "carboxyl"],
                """Number of C-termini

                    Integer 0 (default) or more.
                    """,
            ),
            _Option(
                ["-lysinemodified", "lysinemodified"],
                """Number of modified lysines

                    Integer 0 (default) or more.
                    """,
            ),
            _Option(
                ["-disulphides", "disulphides"],
                """Number of disulphide bridges

                    Integer 0 (default) or more.
                    """,
            ),
            # Should we implement the -termini switch as well?
            _Option(
                ["-notermini", "notermini"],
                "Exclude (True) or include (False) charge at N and C terminus.",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


# seqret uses -outseq, not -outfile, so use the base class:
class SeqretCommandline(_EmbossMinimalCommandLine):
    """Commandline object for the seqret program from EMBOSS.

    This tool allows you to interconvert between different sequence file
    formats (e.g. GenBank to FASTA). Combining Biopython's Bio.SeqIO module
    with seqret using a suitable intermediate file format can allow you to
    read/write to an even wider range of file formats.

    This wrapper currently only supports the core functionality, things like
    feature tables (in EMBOSS 6.1.0 onwards) are not yet included.
    """

    def __init__(self, cmd="seqret", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"], "Input sequence(s) filename", filename=True
            ),
            _Option(["-outseq", "outseq"], "Output sequence file.", filename=True),
            _Option(
                ["-sformat", "sformat"],
                "Input sequence(s) format (e.g. fasta, genbank)",
            ),
            _Option(
                ["-osformat", "osformat"],
                "Output sequence(s) format (e.g. fasta, genbank)",
            ),
        ]
        _EmbossMinimalCommandLine.__init__(self, cmd, **kwargs)

    def _validate(self):
        # Check the outfile, filter, or stdout option has been set.
        # We can't simply do this via the required flag for the outfile
        # output - this seems the simplest solution.
        if not (self.outseq or self.filter or self.stdout):
            raise ValueError(
                "You must either set outfile (output filename), "
                "or enable filter or stdout (output to stdout)."
            )
        if not (self.sequence or self.filter or self.stdint):
            raise ValueError(
                "You must either set sequence (input filename), "
                "or enable filter or stdin (input from stdin)."
            )
        return _EmbossMinimalCommandLine._validate(self)


class SeqmatchallCommandline(_EmbossCommandLine):
    """Commandline object for the seqmatchall program from EMBOSS.

    e.g.
    >>> cline = SeqmatchallCommandline(sequence="opuntia.fasta", outfile="opuntia.txt")
    >>> cline.auto = True
    >>> cline.wordsize = 18
    >>> cline.aformat = "pair"
    >>> print(cline)
    seqmatchall -auto -outfile=opuntia.txt -sequence=opuntia.fasta -wordsize=18 -aformat=pair

    """

    def __init__(self, cmd="seqmatchall", **kwargs):
        """Initialize the class."""
        self.parameters = [
            _Option(
                ["-sequence", "sequence"],
                "Readable set of sequences",
                filename=True,
                is_required=True,
            ),
            _Option(
                ["-wordsize", "wordsize"], "Word size (Integer 2 or more, default 4)"
            ),
            _Option(
                ["-aformat", "aformat"],
                "Display output in a different specified output format",
            ),
        ]
        _EmbossCommandLine.__init__(self, cmd, **kwargs)


if __name__ == "__main__":
    from Bio._utils import run_doctest

    run_doctest()



# Copyright (C) 2002, Thomas Hamelryck (thamelry@binf.ku.dk)
#
# This file is part of the Biopython distribution and governed by your
# choice of the "Biopython License Agreement" or the "BSD 3-Clause License".
# Please see the LICENSE file that should have been included as part of this
# package.

"""Turn an mmCIF file into a dictionary."""


from Bio.File import as_handle


class MMCIF2Dict(dict):
    """Parse a mmCIF file and return a dictionary."""

    def __init__(self, filename):
        """Parse a mmCIF file and return a dictionary.

        Arguments:
         - file - name of the PDB file OR an open filehandle

        """
        self.quote_chars = ["'", '"']
        self.whitespace_chars = [" ", "\t"]
        with as_handle(filename) as handle:
            loop_flag = False
            key = None
            tokens = self._tokenize(handle)
            try:
                token = next(tokens)
            except StopIteration:
                return  # for Python 3.7 and PEP 479
            self[token[0:5]] = token[5:]
            i = 0
            n = 0
            for token in tokens:
                if token.lower() == "loop_":
                    loop_flag = True
                    keys = []
                    i = 0
                    n = 0
                    continue
                elif loop_flag:
                    # The second condition checks we are in the first column
                    # Some mmCIF files (e.g. 4q9r) have values in later columns
                    # starting with an underscore and we don't want to read
                    # these as keys
                    if token.startswith("_") and (n == 0 or i % n == 0):
                        if i > 0:
                            loop_flag = False
                        else:
                            self[token] = []
                            keys.append(token)
                            n += 1
                            continue
                    else:
                        self[keys[i % n]].append(token)
                        i += 1
                        continue
                if key is None:
                    key = token
                else:
                    self[key] = [token]
                    key = None

    # Private methods

    def _splitline(self, line):
        # See https://www.iucr.org/resources/cif/spec/version1.1/cifsyntax for the syntax
        in_token = False
        # quote character of the currently open quote, or None if no quote open
        quote_open_char = None
        start_i = 0
        for (i, c) in enumerate(line):
            if c in self.whitespace_chars:
                if in_token and not quote_open_char:
                    in_token = False
                    yield line[start_i:i]
            elif c in self.quote_chars:
                if not quote_open_char and not in_token:
                    quote_open_char = c
                    in_token = True
                    start_i = i + 1
                elif c == quote_open_char and (
                    i + 1 == len(line) or line[i + 1] in self.whitespace_chars
                ):
                    quote_open_char = None
                    in_token = False
                    yield line[start_i:i]
            elif c == "#" and not in_token:
                # Skip comments. "#" is a valid non-comment char inside of a
                # quote and inside of an unquoted token (!?!?), so we need to
                # check that the current char is not in a token.
                return
            elif not in_token:
                in_token = True
                start_i = i
        if in_token:
            yield line[start_i:]
        if quote_open_char:
            raise ValueError("Line ended with quote open: " + line)

    def _tokenize(self, handle):
        empty = True
        for line in handle:
            empty = False
            if line.startswith("#"):
                continue
            elif line.startswith(";"):
                # The spec says that leading whitespace on each line must be
                # preserved while trailing whitespace may be stripped.  The
                # trailing newline must be stripped.
                token_buffer = [line[1:].rstrip()]
                for line in handle:
                    line = line.rstrip()
                    if line.startswith(";"):
                        yield "\n".join(token_buffer)
                        line = line[1:]
                        if line and not line[0] in self.whitespace_chars:
                            raise ValueError("Missing whitespace")
                        break
                    token_buffer.append(line)
                else:
                    raise ValueError("Missing closing semicolon")
            yield from self._splitline(line.strip())
        if empty:
            raise ValueError("Empty file.")
